import base64
import logging
import time
from datetime import datetime, timedelta
from logging import Logger
from threading import Lock, Condition, Thread

from mw_helper import ensure_list
from xml.etree import cElementTree as eTree

from messagebus import (
    MessageBus,
    TokenPriority,
    TokenCreator,
    Event,
)
from production.box import ProductionBox
from production.command import (
    ProductionCommandProcessor,
    ToggleTagLinesCommand,
    ChangeProdStateCommand,
    UndoProductionCommand,
    AddTagLinesCommand,
    OrderModifiedCommand,
    PublishOrderCommand,
    ProductionCommand,
    PurgeOrdersCommand,
    DevicesUpdatedCommand,
)
from production.command.undo import (
    UndoCommand,
    TagLinesUndoCommand,
    ChangeProdStateUndoCommand,
    TagLinesUndoCommandType,
)
from production.model import OrderChanger
from production.model import (
    ProductionOrder,
    ProdStates,
)
from production.repository import ProductionRepository
from production.util import handle_datetime_str
from production.view import OrderXml
from typing import (
    List,
    Dict,
    Any,
    Optional,
    Callable,
)
from ._CommandQueue import CommandQueue
from ._PublishScheduler import PublishScheduler

performance_event_logger = logging.getLogger("ProductionPerformanceCommand")


MSGGRP_KDS = '0B'
TK_KDS_UPDATE_STATISTICS = TokenCreator.create_token(TokenPriority.high, MSGGRP_KDS, "21")


class ProductionManager(ProductionCommandProcessor, PublishScheduler, OrderChanger):
    def __init__(
            self,
            production_repository,  # type: ProductionRepository
            purge_interval,         # type: int
            orders_life_time,       # type: int
            message_bus,            # type: MessageBus
            audit_prod_states,      # type: List
            kds_manager,            # type: str
            logger                  # type: Logger
    ):
        # type: (...) -> None

        self.production_repository = production_repository
        self.purge_interval = purge_interval
        self.orders_life_time = orders_life_time
        self.message_bus = message_bus
        self.logger = logger
        self.audit_logger_prod_states = audit_prod_states
        self.kds_manager = kds_manager

        self.boxes = []  # type: List[ProductionBox]
        self.orders = {}  # type: Dict[str, ProductionOrder]
        self.undo_lock = Lock()
        self.undo_view_map = {}  # type: Dict[str, List[UndoCommand]]
        self.waiters = {}  # type: Dict[str, ProductionManager.PublishWaiter]
        self.waiter_lock = Lock()
        self.purge_waiter = None  # type: Optional[ProductionManager.PurgeWaiter]
        self.statistics = {}  # type: Dict[str, Any]
        self.view_boxes = {}

        self.command_queue = CommandQueue(self, self.logger)

    def remove_tags(self, items):
        for item in items:
            item.tags = {}
            item.tag_history = []

            self.remove_tags(item.items)

    def initialize(self):
        self.purge_waiter = ProductionManager.PurgeWaiter(self, self.logger)
        self.purge_waiter.purge_orders_by_life_time()
        self._wait_purging_orders()

        init_orders = self.production_repository.get_all_orders()
        self.logger.info("Loading {} ORDERS from Database... Please Wait...".format(len(init_orders)))

        for order in init_orders:
            order.first_processing = False
            for undo_command in order.undo_list:  # type: UndoCommand
                self.enqueue_undo_command(undo_command.view, undo_command)
            self.orders[order.order_id] = order
            self.sync_publish_order(order.order_id)

        self.purge_waiter.purge_orders_and_wait()

    def purge_orders(
        self,           # type: ProductionManager
        order_ids,      # type: List[str]
    ):
        # type: (...) -> None

        self.purge_waiter = ProductionManager.PurgeWaiter(self, self.logger)
        self.purge_waiter.purge_orders_by_id(
            order_ids=order_ids
        )

    @staticmethod
    def _wait_purging_orders():
        time.sleep(5)

    def set_root_boxes(self, boxes):
        # type: (List[ProductionBox]) -> None
        self.boxes = boxes
        for box in self.boxes:
            box.parent = self

    def set_views(self, view_boxes):
        # type: (List) -> None

        for box in view_boxes:
            self.view_boxes[box.get_view_name()] = box

    def process_command(
        self,
        command,        # type: ProductionCommand
        queue_size=0,   # type: Optional[int]
    ):
        # type: (...) -> None

        command_name = type(command).__name__
        order_id = "-"
        try:
            start_time = datetime.now()
            tree_box_count = self.inner_process_command(command)

            if command.order_id is not None:
                order_id = command.order_id.split("-")[0]

            self._performance_log(
                command=command,
                command_name=command_name,
                order_id=order_id,
                start_time=start_time,
                tree_box_count=tree_box_count,
                queue_size=queue_size,
            )
        except Exception as ex:
            performance_event_logger.exception("Error executing command {} for order {}".format(command_name, order_id))
            raise ex

    @staticmethod
    def _performance_log(
        command,            # type: Any
        command_name,       # type: str
        order_id,           # type: int
        start_time,         # type: datetime
        tree_box_count,     # type: Any[Dict, Any]
        queue_size,         # type: int
    ):
        # type: (...) -> None

        end_time = datetime.now()
        time_delta = (end_time - start_time).total_seconds()
        message = "Queue: {:2} -- {:>25} / {:>10} / {:>8.3f}s".format(queue_size, command_name, order_id, time_delta)
        if hasattr(command, 'order') and command.order is not None:
            last_order_update_state_str = handle_datetime_str(command.order.state_history[-1].timestamp)
            last_order_update_state = datetime.strptime(last_order_update_state_str, "%Y-%m-%dT%H:%M:%S.%f")
            last_order_update_state_difference = (start_time - last_order_update_state).total_seconds()

            if isinstance(command, OrderModifiedCommand):
                message += " - Info: {:7} / {:>12} / {:>12} / {} / {} / {:>8.3f}s".format(
                    command.order.originator,
                    command.order.state,
                    command.order.prod_state,
                    command.order.recalled,
                    command.order.state_history[-1].timestamp,
                    last_order_update_state_difference,
                )

        if tree_box_count is not None:
            tree_messages = []
            total_boxes = 0
            for key, value in tree_box_count.iteritems():
                tree_messages.append("{} - {:3}".format(key, str(value)))
                total_boxes += value

            message += " / " + " / ".join(tree_messages) + " / " + str(total_boxes)

        performance_event_logger.info(message)

    def inner_process_command(self, command):
        # type: (ProductionCommand) -> Optional[Dict]

        order = None
        if command.order_id is not None:
            command.order_id = command.order_id.split("-")[0]

        if isinstance(command, OrderModifiedCommand):
            self._handle_order_modified_command(command)
            return

        elif isinstance(command, PublishOrderCommand):
            tree_box_count = self.sync_publish_order(command.order_id)
            return tree_box_count

        elif isinstance(command, PurgeOrdersCommand):
            self.handle_purge_orders_command(command=command)
            return

        if command.order_id is not None and command.order_id not in self.orders:
            if command.view in self.view_boxes:
                current_view = self.view_boxes[command.view]
                current_view.delete_order(command.order_id)
                current_view.reload_view()

            return

        if isinstance(command, ToggleTagLinesCommand) or isinstance(command, AddTagLinesCommand):
            order = self.tag_lines(command)

        elif isinstance(command, ChangeProdStateCommand):
            order = self._handle_change_prod_state_command(command)
            if order is not None:
                self.publish_audit_logger_event(order)

        elif isinstance(command, UndoProductionCommand):
            order = self._handle_undo_production_command(command)

        elif isinstance(command, DevicesUpdatedCommand):
            self._handle_devices_updated()

        else:
            self.logger.error("[process_command] Unknown command received: {}".format(type(command)))

        if order is not None:
            self.production_repository.save_order(order)
            self.publish_order(order.order_id)

    def _handle_undo_production_command(self, command):
        try:
            view_undo_commands = self.undo_view_map[command.view]
            if len(view_undo_commands) == 0:
                return
        except KeyError:
            self.logger.info("No Undo Commands for the View: {}".format(command.view))
            return

        with self.undo_lock:
            next_available = None
            for undo_command in view_undo_commands:
                if undo_command.is_enabled():
                    next_available = undo_command
                    break

            if next_available is not None:
                view_undo_commands.remove(next_available)

        if next_available is None:
            return

        if next_available.order_id not in self.orders:
            self.logger.info("The order {} has no longer available".format(next_available.order_id))
            return

        order = self.orders[next_available.order_id]
        if order is None:
            return
        next_available.undo(order)

        return order

    def _handle_change_prod_state_command(self, command):
        order = self.orders[command.order_id]
        if order.prod_state == command.state:
            self.logger.debug("Order {} is already in prod state {}".format(command.order_id, command.state))
            return

        state_undo_command = ChangeProdStateUndoCommand(order.order_id, command.view, order.prod_state, [])
        self.add_undo_command_to_order_and_view(command.view, order, state_undo_command)

        order.prod_state = command.state
        return order

    def handle_purge_orders_command(
        self,
        command,    # type: PurgeOrdersCommand
    ):
        # type: (...) -> None

        self.logger.debug("Purging orders command start")
        self.production_repository.purge_orders(command.orders)
        cached_purged_orders = self._get_cached_purged_orders(command.orders)
        self.logger.debug("Syncing cached purged orders")
        for order in cached_purged_orders:
            self.sync_publish_order(order.order_id)
            self.orders.pop(order.order_id, None)

    def _handle_order_modified_command(self, command):
        order = command.order
        if order.order_id in self.orders:
            saved_order = self.orders[order.order_id]
            if len(saved_order.state_history) > len(order.state_history):
                self.logger.info("Order Id {} ignored for state history".format(order.order_id))
                return

            order.tags = saved_order.tags
            order.prod_state = saved_order.prod_state
            order.selected_pickup_area = saved_order.selected_pickup_area
            order.first_processing = saved_order.first_processing

            self.update_order_items(order, saved_order)
            if hasattr(saved_order, "round_robin_path"):
                order.round_robin_path = saved_order.round_robin_path
        else:
            order.first_processing = True

        self.orders[order.order_id] = order
        self.production_repository.save_order(order)

        self.delayed_publish_order(order.order_id)

    def _handle_devices_updated(self):
        # type: (...) -> None

        self.logger.info("Event DEVICES_UPDATED received, passing to all view boxes")
        for view_box in self.view_boxes.values():
            view_box.handle_devices_status_changed()

    def order_modified(self, order):
        # type: (ProductionOrder) -> None

        self.command_queue.put(OrderModifiedCommand(order.order_id, order))

    def update_order_items(self, order, saved_order):
        # type: (ProductionOrder, ProductionOrder) -> None
        for item in order.items:
            for saved_item in saved_order.items:
                if item.get_line_id() == saved_item.get_line_id():
                    self.add_saved_tags_to_item(item, saved_item)

    def add_saved_tags_to_item(self, item, saved_item):
        item.tags = saved_item.tags
        item.tag_history = saved_item.tag_history
        item.jit_lines = saved_item.jit_lines
        for son in item.items:
            for saved_son in saved_item.items:
                if son.get_line_id() == saved_son.get_line_id():
                    self.add_saved_tags_to_item(son, saved_son)

    def handle_command(self, command):
        self.command_queue.put(command)

    def schedule_publish(self, order_id, wait_time):
        if order_id is not None:
            order_id = str(order_id)
            order_id = order_id.split("-")[0]

        if wait_time >= self.orders_life_time * 60:
            return

        wait_time = max(wait_time, 0)
        publish_waiter = ProductionManager.PublishWaiter(self, order_id, wait_time)
        waiter_to_deactivate = None
        with self.waiter_lock:
            if order_id in self.waiters:
                publish_time = min(self.waiters[order_id].date_limit, publish_waiter.date_limit)
                due_limite = (publish_time - datetime.now()).total_seconds()
                message = "Saved a Publish Order on Order ID {} - Will be Publish in {}"
                performance_event_logger.info(message.format(order_id, due_limite))

                if self.waiters[order_id].date_limit > publish_waiter.date_limit:
                    waiter_to_deactivate = self.waiters[order_id]
                    del self.waiters[order_id]
                    self.waiters[order_id] = publish_waiter
                else:
                    return
            else:
                self.waiters[order_id] = publish_waiter

        if waiter_to_deactivate is not None:
            waiter_to_deactivate.deactivate()

        publish_waiter.wait_and_publish()

    def remove_schedule_publish(self, order_id):
        # type: (str) -> None

        if order_id is not None:
            order_id = str(order_id)
            order_id = order_id.split("-")[0]

        if order_id in self.waiters:
            self.waiters[order_id].deactivate()
            del self.waiters[order_id]

    def change_order(self, order_id, order_changer):
        # type: (str, Callable[[ProductionOrder], bool]) -> None

        if order_id is not None:
            order_id = str(order_id)

        order_id = order_id.split("-")[0]

        if order_id in self.orders:
            order = self.orders[order_id]
            changed_order = order_changer(order)

            if changed_order:
                self.production_repository.save_order(order)
                self.publish_order(order.order_id)

    def set_all_items(self, item_view, item_save, item_setter):
        def get_diff_key(x):
            return '{}-{}-{}'.format(x.line_number, x.item_id, x.part_code)

        if item_save.is_product():
            if get_diff_key(item_save) == get_diff_key(item_view):
                return item_setter(item_save)
            return False

        were_change = False
        for son in item_save.items:
            if self.set_all_items(item_view, son, item_setter):
                were_change = True

        return were_change

    def change_items(self, order, item_changer):
        order_id = order.order_id.split("-")[0]
        saved_order = self.orders[order_id]
        was_change = False

        for view_item in order.items:
            for saved_item in saved_order.items:
                if self.set_all_items(view_item, saved_item, item_changer):
                    was_change = True

        if was_change:
            self.production_repository.save_order(saved_order)

    def tag_lines(self, command):
        order_id = command.order_id.split("-")[0]
        order = self.orders[order_id]
        self.change_order_tags(command, order)
        self.create_undo_command(command, order)
        return order

    @staticmethod
    def change_order_tags(command, order):
        tags = ensure_list(command.tag)

        if isinstance(command, ToggleTagLinesCommand):
            command_callable = order.toggle_line_tag
        else:
            command_callable = order.add_line_tag

        for tag in tags:
            for line_id in command.line_ids:
                line_jit_list = command.line_ids_jit_lines.get(line_id) if command.line_ids_jit_lines else None
                command_callable(line_id, tag, line_jit_list)

    def create_undo_command(self, command, order):
        undo_command = self.build_undo_command(command, order)
        self.add_undo_command_to_order_and_view(command.view, order, undo_command)

    def build_undo_command(self, command, order):
        disabled_commands = self.get_disabled_commands(command, order)
        if isinstance(command, ToggleTagLinesCommand):
            undo_type = TagLinesUndoCommandType.toggle
        else:
            undo_type = TagLinesUndoCommandType.add

        undo_command = TagLinesUndoCommand(
            order.order_id,
            command.view,
            command.tag,
            command.line_ids,
            undo_type,
            disabled_commands,
            command.line_ids_jit_lines
        )

        return undo_command

    @staticmethod
    def get_disabled_commands(command, order):
        disabled_commands = []
        if command.tag == "served":
            for undo_command in order.undo_list:
                if isinstance(undo_command, TagLinesUndoCommand) and undo_command.tag in ("doing", "done"):
                    undo_command.disable()
                    disabled_commands.append(undo_command)
        return disabled_commands

    def get_order_xml(self, order_id):
        if order_id not in self.orders:
            return None

        return OrderXml().to_xml(self.orders[order_id])

    def get_production_order(self, order_id):
        if order_id not in self.orders:
            return None

        return self.orders[order_id]

    def publish_order(self, order_id):
        if order_id not in self.orders:
            return

        self.command_queue.put(PublishOrderCommand(order_id))

    def delayed_publish_order(self, order_id):
        self.schedule_publish(order_id, 1)

    def sync_publish_order(self, order_id):
        with self.waiter_lock:
            if order_id in self.waiters:
                self.waiters[order_id].deactivate()
                del self.waiters[order_id]

        if order_id not in self.orders:
            self.logger.error("OrderId {} not in self.orders".format(order_id))
            return {}

        self.logger.debug(">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>", extra={'simple': True})
        self.logger.debug("#region    - Publishing order: {}".format(order_id), extra={'simple': True})

        order = self.orders[order_id]
        tree_box_count = {}
        for box in self.boxes:
            try:
                self.logger.debug("#region    - {} @ {}".format(box.name, box.tree_name), extra={'simple': True})
                cloned_order = order.clone()
                box.order_modified(cloned_order)
                tree_box_count[box.tree_name] = cloned_order.box_count if hasattr(cloned_order, 'box_count') else 1
            except Exception as _:  # noqa
                self.logger.exception("Exception on box: {}".format(box.name))
            finally:
                self.logger.debug("#endregion - {} @ {}".format(box.name, box.tree_name), extra={'simple': True})

        final_message = "#endregion    - Finished publishing order: {}".format(order.order_id)
        self.logger.debug(final_message, extra={'simple': True})
        self.logger.debug("<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<", extra={'simple': True})

        order.first_processing = False
        return tree_box_count

    def publish_command(self, command):
        handled = []
        for box in self.boxes:
            try:
                if box.handle_production_command(command):
                    handled.append(box)
            except Exception as _:  # noqa
                self.logger.exception("Exception on box: {}".format(box.name))

        return handled

    @staticmethod
    def add_tag_to_dict(tags, line_id, tag):
        if line_id not in tags:
            tags[line_id] = []
        tags[line_id].append(tag)

    def add_undo_command_to_order_and_view(self, view, order, undo_command):
        order.undo_list.append(undo_command)
        self.enqueue_undo_command(view, undo_command)

    def enqueue_undo_command(self, view, undo_command):
        with self.undo_lock:
            if view not in self.undo_view_map:
                self.undo_view_map[view] = []
            self.undo_view_map[view].insert(0, undo_command)

    def stop(self):
        if self.purge_waiter is not None:
            self.purge_waiter.deactivate()

    def _get_cached_purged_orders(self, orders_to_purge):
        # type: (List[str]) -> List[ProductionOrder]
        orders = []
        for order_id in orders_to_purge:
            str_order_id = str(order_id)
            if str_order_id in self.orders:
                order = self.orders[str_order_id]
                order.purged = True
                order.prod_state = ProdStates.INVALID
                orders.append(order)
        return orders

    def publish_audit_logger_event(self, order):
        if order.prod_state not in self.audit_logger_prod_states:
            return

        order_id = str(order.order_id)
        try:
            production_order = eTree.fromstring(ProductionManager.get_order_xml(self, order.order_id))
            production_order_xml = eTree.tostring(production_order)

            data = """<ProductionOrderEvent posId="{0}" businessPeriod="{1}" orderId="{2}">{3}</ProductionOrderEvent>"""
            msg_data = data.format(str(order.pos_id), str(order.business_day), order_id, production_order_xml)
            event = Event(subject="PRODUCTION_ORDER_EVENT", event_type="PRODUCTION_ORDER", data=msg_data)

            self.logger.info("Publishing AuditLog event: {} / {}".format(event.subject, event.event_type))
            self.message_bus.publish_event(event)
        except Exception as _:  # noqa
            self.logger.exception("Error publishing AuditLogger event for orderId#{}".format(order_id))

    class PublishWaiter(object):
        def __init__(self, production_manager, order_id, wait_time):
            self.production_manager = production_manager
            self.order_id = order_id
            self.wait_time = wait_time
            self.date_limit = datetime.now() + timedelta(seconds=self.wait_time)
            self.condition = Condition()
            self.active = True

        def wait_and_publish(self):
            t = Thread(target=self._wait_and_publish, name="PublishWaiter")
            t.daemon = True
            t.start()

        def deactivate(self):
            self.active = False
            with self.condition:
                self.condition.notify_all()

        def _wait_and_publish(self):
            with self.condition:
                self.condition.wait(self.wait_time)

            if self.active:
                self.production_manager.publish_order(self.order_id)

    class PurgeWaiter(object):
        def __init__(self, production_manager, logger):
            self.production_manager = production_manager
            self.logger = logger
            self.purge_interval = production_manager.purge_interval
            self.orders_life_time = production_manager.orders_life_time
            self.production_repository = production_manager.production_repository
            self.orders = production_manager.orders
            self.condition = Condition()
            self.active = True

        def deactivate(self):
            self.active = False
            with self.condition:
                self.condition.notify_all()

        def purge_orders_and_wait(self):
            t = Thread(target=self._purge_orders_and_wait, name="PurgeWaiter")
            t.daemon = True
            t.start()

        def _purge_orders_and_wait(self):
            while self.active:
                try:
                    self.logger.debug("Starting thread!")
                    self.purge_orders_by_life_time()

                    with self.condition:
                        sleep_time = self.purge_interval * 60
                        self.logger.debug("Sleeping for {} seconds".format(sleep_time))
                        self.condition.wait(sleep_time)
                except (Exception,):
                    self.logger.exception("Error purging old orders")

        def purge_orders_by_life_time(
            self
        ):
            # type: (...) -> None

            self.logger.info("Starting purge process by orders life time")

            limit_date = datetime.now() - timedelta(minutes=self.orders_life_time)
            orders_to_purge = self.production_repository.get_orders_to_purge(limit_date)
            self._purge_orders(
                orders_to_purge=orders_to_purge,
            )

        def purge_orders_by_id(
            self,
            order_ids=None,         # type: Optional[List[str]]
        ):
            # type: (...) -> None

            self.logger.info("Starting purge process by order id: {}".format(",".join(order_ids)))

            if order_ids is not None and len(order_ids) > 0:
                orders_to_purge = self.production_repository.get_orders_by_orders_ids(order_ids)
                orders_to_purge = [x.order_id for x in orders_to_purge]
            else:
                all_orders = self.production_repository.get_all_orders()
                orders_to_purge = [x.order_id for x in all_orders]

            self._purge_orders(
                orders_to_purge=orders_to_purge,
            )

        def _purge_orders(
            self,
            orders_to_purge,        # type: List[str]
        ):
            # type: (...) -> None

            self.logger.info("Found {} orders to purge".format(len(orders_to_purge)))
            if len(orders_to_purge) != 0:
                self.production_manager.handle_command(PurgeOrdersCommand(orders_to_purge))

    def save_pick_list_order(
        self,
        order,          # type: ProductionOrder
        receipt,        # type: str
        status,         # type: bool
    ):
        # type: (...) -> None

        order_id = order.order_id.split('-')[0]
        order_id_count = order.order_id if '-' in order.order_id else ''

        receipt_data = base64.b64encode(
            s=receipt,
        )

        self.production_repository.save_pick_list_order(
            order_id=order_id,
            order_id_count=order_id_count,
            jit=order.jit_printer_name,
            course=order.current_course,
            receipt=receipt_data,
            status=status,
        )

    def get_picklist_id(
        self,
        order,      # type: ProductionOrder
    ):
        # type: (...) -> Optional[int]

        order_id = order.order_id.split('-')[0]
        order_id_count = order.order_id if '-' in order.order_id else ''

        return self.production_repository.get_picklist_id(
            order_id=order_id,
            order_id_count=order_id_count,
            jit=order.jit_printer_name,
            course=order.current_course,
        )



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from logging import getLogger
from threading import RLock
from time import time_ns
from typing import Dict, List, Optional

from opentelemetry.metrics import (
    Asynchronous,
    Counter,
    Instrument,
    ObservableCounter,
)
from opentelemetry.sdk.metrics._internal._view_instrument_match import (
    _ViewInstrumentMatch,
)
from opentelemetry.sdk.metrics._internal.aggregation import (
    Aggregation,
    ExplicitBucketHistogramAggregation,
    _DropAggregation,
    _ExplicitBucketHistogramAggregation,
    _ExponentialBucketHistogramAggregation,
    _LastValueAggregation,
    _SumAggregation,
)
from opentelemetry.sdk.metrics._internal.export import AggregationTemporality
from opentelemetry.sdk.metrics._internal.measurement import Measurement
from opentelemetry.sdk.metrics._internal.point import (
    ExponentialHistogram,
    Gauge,
    Histogram,
    Metric,
    MetricsData,
    ResourceMetrics,
    ScopeMetrics,
    Sum,
)
from opentelemetry.sdk.metrics._internal.sdk_configuration import (
    SdkConfiguration,
)
from opentelemetry.sdk.metrics._internal.view import View
from opentelemetry.sdk.util.instrumentation import InstrumentationScope

_logger = getLogger(__name__)

_DEFAULT_VIEW = View(instrument_name="")


class MetricReaderStorage:
    """The SDK's storage for a given reader"""

    def __init__(
        self,
        sdk_config: SdkConfiguration,
        instrument_class_temporality: Dict[type, AggregationTemporality],
        instrument_class_aggregation: Dict[type, Aggregation],
    ) -> None:
        self._lock = RLock()
        self._sdk_config = sdk_config
        self._instrument_view_instrument_matches: Dict[
            Instrument, List[_ViewInstrumentMatch]
        ] = {}
        self._instrument_class_temporality = instrument_class_temporality
        self._instrument_class_aggregation = instrument_class_aggregation

    def _get_or_init_view_instrument_match(
        self, instrument: Instrument
    ) -> List[_ViewInstrumentMatch]:
        # Optimistically get the relevant views for the given instrument. Once set for a given
        # instrument, the mapping will never change

        if instrument in self._instrument_view_instrument_matches:
            return self._instrument_view_instrument_matches[instrument]

        with self._lock:
            # double check if it was set before we held the lock
            if instrument in self._instrument_view_instrument_matches:
                return self._instrument_view_instrument_matches[instrument]

            # not present, hold the lock and add a new mapping
            view_instrument_matches = []

            self._handle_view_instrument_match(
                instrument, view_instrument_matches
            )

            # if no view targeted the instrument, use the default
            if not view_instrument_matches:
                view_instrument_matches.append(
                    _ViewInstrumentMatch(
                        view=_DEFAULT_VIEW,
                        instrument=instrument,
                        instrument_class_aggregation=(
                            self._instrument_class_aggregation
                        ),
                    )
                )
            self._instrument_view_instrument_matches[instrument] = (
                view_instrument_matches
            )

            return view_instrument_matches

    def consume_measurement(
        self, measurement: Measurement, should_sample_exemplar: bool = True
    ) -> None:
        for view_instrument_match in self._get_or_init_view_instrument_match(
            measurement.instrument
        ):
            view_instrument_match.consume_measurement(
                measurement, should_sample_exemplar
            )

    def collect(self) -> Optional[MetricsData]:
        # Use a list instead of yielding to prevent a slow reader from holding
        # SDK locks

        # While holding the lock, new _ViewInstrumentMatch can't be added from
        # another thread (so we are sure we collect all existing view).
        # However, instruments can still send measurements that will make it
        # into the individual aggregations; collection will acquire those locks
        # iteratively to keep locking as fine-grained as possible. One side
        # effect is that end times can be slightly skewed among the metric
        # streams produced by the SDK, but we still align the output timestamps
        # for a single instrument.

        collection_start_nanos = time_ns()

        with self._lock:
            instrumentation_scope_scope_metrics: Dict[
                InstrumentationScope, ScopeMetrics
            ] = {}

            for (
                instrument,
                view_instrument_matches,
            ) in self._instrument_view_instrument_matches.items():
                aggregation_temporality = self._instrument_class_temporality[
                    instrument.__class__
                ]

                metrics: List[Metric] = []

                for view_instrument_match in view_instrument_matches:
                    data_points = view_instrument_match.collect(
                        aggregation_temporality, collection_start_nanos
                    )

                    if data_points is None:
                        continue

                    if isinstance(
                        # pylint: disable=protected-access
                        view_instrument_match._aggregation,
                        _SumAggregation,
                    ):
                        data = Sum(
                            aggregation_temporality=aggregation_temporality,
                            data_points=data_points,
                            is_monotonic=isinstance(
                                instrument, (Counter, ObservableCounter)
                            ),
                        )
                    elif isinstance(
                        # pylint: disable=protected-access
                        view_instrument_match._aggregation,
                        _LastValueAggregation,
                    ):
                        data = Gauge(data_points=data_points)
                    elif isinstance(
                        # pylint: disable=protected-access
                        view_instrument_match._aggregation,
                        _ExplicitBucketHistogramAggregation,
                    ):
                        data = Histogram(
                            data_points=data_points,
                            aggregation_temporality=aggregation_temporality,
                        )
                    elif isinstance(
                        # pylint: disable=protected-access
                        view_instrument_match._aggregation,
                        _DropAggregation,
                    ):
                        continue

                    elif isinstance(
                        # pylint: disable=protected-access
                        view_instrument_match._aggregation,
                        _ExponentialBucketHistogramAggregation,
                    ):
                        data = ExponentialHistogram(
                            data_points=data_points,
                            aggregation_temporality=aggregation_temporality,
                        )

                    metrics.append(
                        Metric(
                            # pylint: disable=protected-access
                            # pylint: disable=possibly-used-before-assignment
                            name=view_instrument_match._name,
                            description=view_instrument_match._description,
                            unit=view_instrument_match._instrument.unit,
                            data=data,
                        )
                    )

                if metrics:
                    if instrument.instrumentation_scope not in (
                        instrumentation_scope_scope_metrics
                    ):
                        instrumentation_scope_scope_metrics[
                            instrument.instrumentation_scope
                        ] = ScopeMetrics(
                            scope=instrument.instrumentation_scope,
                            metrics=metrics,
                            schema_url=instrument.instrumentation_scope.schema_url,
                        )
                    else:
                        instrumentation_scope_scope_metrics[
                            instrument.instrumentation_scope
                        ].metrics.extend(metrics)

            if instrumentation_scope_scope_metrics:
                return MetricsData(
                    resource_metrics=[
                        ResourceMetrics(
                            resource=self._sdk_config.resource,
                            scope_metrics=list(
                                instrumentation_scope_scope_metrics.values()
                            ),
                            schema_url=self._sdk_config.resource.schema_url,
                        )
                    ]
                )

            return None

    def _handle_view_instrument_match(
        self,
        instrument: Instrument,
        view_instrument_matches: List["_ViewInstrumentMatch"],
    ) -> None:
        for view in self._sdk_config.views:
            # pylint: disable=protected-access
            if not view._match(instrument):
                continue

            if not self._check_view_instrument_compatibility(view, instrument):
                continue

            new_view_instrument_match = _ViewInstrumentMatch(
                view=view,
                instrument=instrument,
                instrument_class_aggregation=(
                    self._instrument_class_aggregation
                ),
            )

            for (
                existing_view_instrument_matches
            ) in self._instrument_view_instrument_matches.values():
                for (
                    existing_view_instrument_match
                ) in existing_view_instrument_matches:
                    if existing_view_instrument_match.conflicts(
                        new_view_instrument_match
                    ):
                        _logger.warning(
                            "Views %s and %s will cause conflicting "
                            "metrics identities",
                            existing_view_instrument_match._view,
                            new_view_instrument_match._view,
                        )

            view_instrument_matches.append(new_view_instrument_match)

    @staticmethod
    def _check_view_instrument_compatibility(
        view: View, instrument: Instrument
    ) -> bool:
        """
        Checks if a view and an instrument are compatible.

        Returns `true` if they are compatible and a `_ViewInstrumentMatch`
        object should be created, `false` otherwise.
        """

        result = True

        # pylint: disable=protected-access
        if isinstance(instrument, Asynchronous) and isinstance(
            view._aggregation, ExplicitBucketHistogramAggregation
        ):
            _logger.warning(
                "View %s and instrument %s will produce "
                "semantic errors when matched, the view "
                "has not been applied.",
                view,
                instrument,
            )
            result = False

        return result



from logging import Logger
from typing import (
    Optional,
    List,
)

from production.model import (
    ProductionOrder,
    Item,
)
from .._OrderChangerProductionBox import OrderChangerProductionBox


class StoredOrderUnsplitter(OrderChangerProductionBox):

    def __init__(
        self,                       # type: StoredOrderUnsplitter
        name,                       # type: str
        sons,                       # type: object
        prod_states_clean_cache,    # type: List[str]
        logger=None,                # type: Optional[Logger]
    ):
        # type: (...) -> None

        super(StoredOrderUnsplitter, self).__init__(name, sons, logger)

        self.prod_states_clean_cache = prod_states_clean_cache

    def change_order(
        self,     # type: StoredOrderUnsplitter
        order,    # type: ProductionOrder
    ):
        # type: (...) -> ProductionOrder

        if not order.stored_split_order:
            return order

        self._update_cached_order(
            new_order=order
        )

        return self._get_merged_order(order=order)

    def _update_cached_order(
        self,            # type: StoredOrderUnsplitter
        new_order,       # type: ProductionOrder
    ):
        # type: (...) -> None

        if new_order.prod_state in self.prod_states_clean_cache:
            self.delete_order(order_id=new_order.order_id)
            return

        cached_order = self.get_order(order_id=new_order.order_id)
        if cached_order is None:
            self.save_order(order=new_order)
            return

        new_order.items = self._get_merged_items(
            new_order=new_order,
            cache_order=cached_order,
        )

        self.save_order(order=new_order)

    @staticmethod
    def _get_merged_items(
        new_order,      # type: ProductionOrder
        cache_order,    # type: ProductionOrder
    ):
        # type: (...) ->  List[Item]

        merged_items = cache_order.items

        items_to_remove_from_cache = []
        for cloned_item in new_order.items:
            for cached_item in merged_items:
                if cloned_item.get_line_id() == cached_item.get_line_id():
                    items_to_remove_from_cache.append(cached_item)
                    break

        for item_to_remove in items_to_remove_from_cache:
            merged_items.remove(item_to_remove)

        merged_items.extend(new_order.items)
        return merged_items

    def _get_merged_order(
        self,       # type: StoredOrderUnsplitter
        order,      # type: ProductionOrder
    ):
        # type: (...) -> ProductionOrder

        order_cloned = order.clone(items=order.items)
        split_indexes = order.order_id.split("-")
        order_cloned.order_id = "-".join(split_indexes[:len(split_indexes) - 1])

        for index in range(0, 9):
            temp_order = self.get_order(order_id="-".join([order_cloned.order_id, str(index)]))
            if temp_order:
                order_cloned.items = self._get_merged_items(
                    new_order=temp_order,
                    cache_order=order_cloned,
                )

        order_cloned.stored_split_order = False
        return order_cloned



from logging import Logger

import iso8601
from production.model import (
    ProductionOrder,
    StateEvent,
    Item,
)
from typing import (
    List,
    Optional,
    Set,
)
from .._OrderChangerProductionBox import OrderChangerProductionBox


class StoredOrderSplitter(OrderChangerProductionBox):

    def __init__(
        self,           # type: StoredOrderSplitter
        name,           # type: str
        sons,           # type: object
        logger=None,    # type: Optional[Logger]
    ):
        # type: (...) -> None

        super(StoredOrderSplitter, self).__init__(name, sons, logger)

    def change_order(
        self,       # type: StoredOrderSplitter
        order,      # type: ProductionOrder
    ):
        # type: (...) -> List[ProductionOrder]

        visited_items = set()
        orders = self._get_splitted_stored_orders(
            order=order,
            visited_items=visited_items,
        )

        self._add_not_visited_items_to_new_order(
            order=order,
            cloned_orders=orders,
            visited_items=visited_items,
        )

        return orders

    def _get_splitted_stored_orders(
        self,             # type: StoredOrderSplitter
        order,            # type: ProductionOrder
        visited_items,    # type: Set[Item]
    ):
        # type: (...) -> List[ProductionOrder]

        orders = []
        order_index = 0
        states_stored = self._get_stored_states(order=order)
        for state in states_stored:
            order_clone = self._clone_order_to_split(
                order=order,
                order_index=order_index,
                new_display_time=state.timestamp,
            )
            state_datetime = iso8601.parse_date(state.timestamp_gmt)
            for item in order.items:
                created_at = iso8601.parse_date(item.created_at)
                if created_at <= state_datetime and item not in visited_items:
                    order_clone.items.append(item)
                    visited_items.add(item)
                    if order_index > 0:
                        item.add_tag(tag="item-new-added")

            if len(order_clone.items) > 0:
                orders.append(order_clone)
                order_index += 1

        return orders

    @staticmethod
    def _clone_order_to_split(
        order,               # type: ProductionOrder
        order_index,         # type: int
        new_display_time,    # type: str
    ):
        # type: (...) -> ProductionOrder

        order_clone = order.clone(
            items=[],
            index=order_index,
        )
        order_clone.display_time = iso8601.parse_date(new_display_time).isoformat()
        order_clone.stored_split_order = True

        return order_clone

    @staticmethod
    def _add_not_visited_items_to_new_order(
        order,              # type: ProductionOrder
        cloned_orders,      # type: List[ProductionOrder]
        visited_items,      # type: Set[Item]
    ):
        # type: (...) -> None

        order_clone = order.clone(
            items=[],
            index=len(cloned_orders),
        )
        order_clone.stored_split_order = True
        for item in order.items:
            if item not in visited_items:
                order_clone.items.append(item)

        if len(order_clone.items) > 0:
            cloned_orders.append(order_clone)

    @staticmethod
    def _get_stored_states(
        order,    # type: ProductionOrder
    ):
        # type: (...) -> List[StateEvent]

        stored_states = []
        for state in order.state_history:
            if state.state == "STORED":
                stored_states.append(state)

        return sorted(stored_states, key=lambda item: item.timestamp_gmt)



#
# The Python Imaging Library
# $Id$
#
# JPEG2000 file handling
#
# History:
# 2014-03-12 ajh  Created
# 2021-06-30 rogermb  Extract dpi information from the 'resc' header box
#
# Copyright (c) 2014 Coriolis Systems Limited
# Copyright (c) 2014 Alastair Houghton
#
# See the README file for information on usage and redistribution.
#
from __future__ import annotations

import io
import os
import struct

from . import Image, ImageFile, _binary


class BoxReader:
    """
    A small helper class to read fields stored in JPEG2000 header boxes
    and to easily step into and read sub-boxes.
    """

    def __init__(self, fp, length=-1):
        self.fp = fp
        self.has_length = length >= 0
        self.length = length
        self.remaining_in_box = -1

    def _can_read(self, num_bytes):
        if self.has_length and self.fp.tell() + num_bytes > self.length:
            # Outside box: ensure we don't read past the known file length
            return False
        if self.remaining_in_box >= 0:
            # Inside box contents: ensure read does not go past box boundaries
            return num_bytes <= self.remaining_in_box
        else:
            return True  # No length known, just read

    def _read_bytes(self, num_bytes):
        if not self._can_read(num_bytes):
            msg = "Not enough data in header"
            raise SyntaxError(msg)

        data = self.fp.read(num_bytes)
        if len(data) < num_bytes:
            msg = f"Expected to read {num_bytes} bytes but only got {len(data)}."
            raise OSError(msg)

        if self.remaining_in_box > 0:
            self.remaining_in_box -= num_bytes
        return data

    def read_fields(self, field_format):
        size = struct.calcsize(field_format)
        data = self._read_bytes(size)
        return struct.unpack(field_format, data)

    def read_boxes(self):
        size = self.remaining_in_box
        data = self._read_bytes(size)
        return BoxReader(io.BytesIO(data), size)

    def has_next_box(self):
        if self.has_length:
            return self.fp.tell() + self.remaining_in_box < self.length
        else:
            return True

    def next_box_type(self):
        # Skip the rest of the box if it has not been read
        if self.remaining_in_box > 0:
            self.fp.seek(self.remaining_in_box, os.SEEK_CUR)
        self.remaining_in_box = -1

        # Read the length and type of the next box
        lbox, tbox = self.read_fields(">I4s")
        if lbox == 1:
            lbox = self.read_fields(">Q")[0]
            hlen = 16
        else:
            hlen = 8

        if lbox < hlen or not self._can_read(lbox - hlen):
            msg = "Invalid header length"
            raise SyntaxError(msg)

        self.remaining_in_box = lbox - hlen
        return tbox


def _parse_codestream(fp):
    """Parse the JPEG 2000 codestream to extract the size and component
    count from the SIZ marker segment, returning a PIL (size, mode) tuple."""

    hdr = fp.read(2)
    lsiz = _binary.i16be(hdr)
    siz = hdr + fp.read(lsiz - 2)
    lsiz, rsiz, xsiz, ysiz, xosiz, yosiz, _, _, _, _, csiz = struct.unpack_from(
        ">HHIIIIIIIIH", siz
    )
    ssiz = [None] * csiz
    xrsiz = [None] * csiz
    yrsiz = [None] * csiz
    for i in range(csiz):
        ssiz[i], xrsiz[i], yrsiz[i] = struct.unpack_from(">BBB", siz, 36 + 3 * i)

    size = (xsiz - xosiz, ysiz - yosiz)
    if csiz == 1:
        if (yrsiz[0] & 0x7F) > 8:
            mode = "I;16"
        else:
            mode = "L"
    elif csiz == 2:
        mode = "LA"
    elif csiz == 3:
        mode = "RGB"
    elif csiz == 4:
        mode = "RGBA"
    else:
        mode = None

    return size, mode


def _res_to_dpi(num, denom, exp):
    """Convert JPEG2000's (numerator, denominator, exponent-base-10) resolution,
    calculated as (num / denom) * 10^exp and stored in dots per meter,
    to floating-point dots per inch."""
    if denom != 0:
        return (254 * num * (10**exp)) / (10000 * denom)


def _parse_jp2_header(fp):
    """Parse the JP2 header box to extract size, component count,
    color space information, and optionally DPI information,
    returning a (size, mode, mimetype, dpi) tuple."""

    # Find the JP2 header box
    reader = BoxReader(fp)
    header = None
    mimetype = None
    while reader.has_next_box():
        tbox = reader.next_box_type()

        if tbox == b"jp2h":
            header = reader.read_boxes()
            break
        elif tbox == b"ftyp":
            if reader.read_fields(">4s")[0] == b"jpx ":
                mimetype = "image/jpx"

    size = None
    mode = None
    bpc = None
    nc = None
    dpi = None  # 2-tuple of DPI info, or None

    while header.has_next_box():
        tbox = header.next_box_type()

        if tbox == b"ihdr":
            height, width, nc, bpc = header.read_fields(">IIHB")
            size = (width, height)
            if nc == 1 and (bpc & 0x7F) > 8:
                mode = "I;16"
            elif nc == 1:
                mode = "L"
            elif nc == 2:
                mode = "LA"
            elif nc == 3:
                mode = "RGB"
            elif nc == 4:
                mode = "RGBA"
        elif tbox == b"res ":
            res = header.read_boxes()
            while res.has_next_box():
                tres = res.next_box_type()
                if tres == b"resc":
                    vrcn, vrcd, hrcn, hrcd, vrce, hrce = res.read_fields(">HHHHBB")
                    hres = _res_to_dpi(hrcn, hrcd, hrce)
                    vres = _res_to_dpi(vrcn, vrcd, vrce)
                    if hres is not None and vres is not None:
                        dpi = (hres, vres)
                    break

    if size is None or mode is None:
        msg = "Malformed JP2 header"
        raise SyntaxError(msg)

    return size, mode, mimetype, dpi


##
# Image plugin for JPEG2000 images.


class Jpeg2KImageFile(ImageFile.ImageFile):
    format = "JPEG2000"
    format_description = "JPEG 2000 (ISO 15444)"

    def _open(self):
        sig = self.fp.read(4)
        if sig == b"\xff\x4f\xff\x51":
            self.codec = "j2k"
            self._size, self._mode = _parse_codestream(self.fp)
        else:
            sig = sig + self.fp.read(8)

            if sig == b"\x00\x00\x00\x0cjP  \x0d\x0a\x87\x0a":
                self.codec = "jp2"
                header = _parse_jp2_header(self.fp)
                self._size, self._mode, self.custom_mimetype, dpi = header
                if dpi is not None:
                    self.info["dpi"] = dpi
                if self.fp.read(12).endswith(b"jp2c\xff\x4f\xff\x51"):
                    self._parse_comment()
            else:
                msg = "not a JPEG 2000 file"
                raise SyntaxError(msg)

        if self.size is None or self.mode is None:
            msg = "unable to determine size/mode"
            raise SyntaxError(msg)

        self._reduce = 0
        self.layers = 0

        fd = -1
        length = -1

        try:
            fd = self.fp.fileno()
            length = os.fstat(fd).st_size
        except Exception:
            fd = -1
            try:
                pos = self.fp.tell()
                self.fp.seek(0, io.SEEK_END)
                length = self.fp.tell()
                self.fp.seek(pos)
            except Exception:
                length = -1

        self.tile = [
            (
                "jpeg2k",
                (0, 0) + self.size,
                0,
                (self.codec, self._reduce, self.layers, fd, length),
            )
        ]

    def _parse_comment(self):
        hdr = self.fp.read(2)
        length = _binary.i16be(hdr)
        self.fp.seek(length - 2, os.SEEK_CUR)

        while True:
            marker = self.fp.read(2)
            if not marker:
                break
            typ = marker[1]
            if typ in (0x90, 0xD9):
                # Start of tile or end of codestream
                break
            hdr = self.fp.read(2)
            length = _binary.i16be(hdr)
            if typ == 0x64:
                # Comment
                self.info["comment"] = self.fp.read(length - 2)[2:]
                break
            else:
                self.fp.seek(length - 2, os.SEEK_CUR)

    @property
    def reduce(self):
        # https://github.com/python-pillow/Pillow/issues/4343 found that the
        # new Image 'reduce' method was shadowed by this plugin's 'reduce'
        # property. This attempts to allow for both scenarios
        return self._reduce or super().reduce

    @reduce.setter
    def reduce(self, value):
        self._reduce = value

    def load(self):
        if self.tile and self._reduce:
            power = 1 << self._reduce
            adjust = power >> 1
            self._size = (
                int((self.size[0] + adjust) / power),
                int((self.size[1] + adjust) / power),
            )

            # Update the reduce and layers settings
            t = self.tile[0]
            t3 = (t[3][0], self._reduce, self.layers, t[3][3], t[3][4])
            self.tile = [(t[0], (0, 0) + self.size, t[2], t3)]

        return ImageFile.ImageFile.load(self)


def _accept(prefix):
    return (
        prefix[:4] == b"\xff\x4f\xff\x51"
        or prefix[:12] == b"\x00\x00\x00\x0cjP  \x0d\x0a\x87\x0a"
    )


# ------------------------------------------------------------
# Save support


def _save(im, fp, filename):
    # Get the keyword arguments
    info = im.encoderinfo

    if filename.endswith(".j2k") or info.get("no_jp2", False):
        kind = "j2k"
    else:
        kind = "jp2"

    offset = info.get("offset", None)
    tile_offset = info.get("tile_offset", None)
    tile_size = info.get("tile_size", None)
    quality_mode = info.get("quality_mode", "rates")
    quality_layers = info.get("quality_layers", None)
    if quality_layers is not None and not (
        isinstance(quality_layers, (list, tuple))
        and all(
            isinstance(quality_layer, (int, float)) for quality_layer in quality_layers
        )
    ):
        msg = "quality_layers must be a sequence of numbers"
        raise ValueError(msg)

    num_resolutions = info.get("num_resolutions", 0)
    cblk_size = info.get("codeblock_size", None)
    precinct_size = info.get("precinct_size", None)
    irreversible = info.get("irreversible", False)
    progression = info.get("progression", "LRCP")
    cinema_mode = info.get("cinema_mode", "no")
    mct = info.get("mct", 0)
    signed = info.get("signed", False)
    comment = info.get("comment")
    if isinstance(comment, str):
        comment = comment.encode()
    plt = info.get("plt", False)

    fd = -1
    if hasattr(fp, "fileno"):
        try:
            fd = fp.fileno()
        except Exception:
            fd = -1

    im.encoderconfig = (
        offset,
        tile_offset,
        tile_size,
        quality_mode,
        quality_layers,
        num_resolutions,
        cblk_size,
        precinct_size,
        irreversible,
        progression,
        cinema_mode,
        mct,
        signed,
        fd,
        comment,
        plt,
    )

    ImageFile._save(im, fp, [("jpeg2k", (0, 0) + im.size, 0, kind)])


# ------------------------------------------------------------
# Registry stuff


Image.register_open(Jpeg2KImageFile.format, Jpeg2KImageFile, _accept)
Image.register_save(Jpeg2KImageFile.format, _save)

Image.register_extensions(
    Jpeg2KImageFile.format, [".jp2", ".j2k", ".jpc", ".jpf", ".jpx", ".j2c"]
)

Image.register_mime(Jpeg2KImageFile.format, "image/jp2")



from ._AuthenticateNfcProcessor import AuthenticateNfcProcessor
from ._AuthenticateProcessor import AuthenticateProcessor
from ._GetInfoProcessor import GetInfoProcessor
from ._GetUsersListProcessor import GetUsersListProcessor
from ._InsertOrUpdateUsersProcessor import InsertOrUpdateUsersProcessor



import base64
import logging
import time

from logging import Logger

from messagebus import MessageBus
from typing import List

from nreports_api import NReports
from production.model import ProductionOrder
from production.service import PrinterService

from ._OrderXml import OrderXml
from ._PrinterView import PrinterView

MAX_RETRY = 10
printer_logger = logging.getLogger("ProductionPrinter")


class ReportsPrinterView(PrinterView):
    def __init__(
        self,                   # type: ReportsPrinterView
        name,                   # type: str
        report_name,            # type: str
        printer_name,           # type: str
        pick_list_name,         # type: str
        production_manager,     # type: ProductionManager
        timeout,                # type: int
        message_bus,            # type: MessageBus
        tags_on_print=None,     # type: List[str]
        logger=None,            # type: Logger
    ):
        # type: (...) -> None

        super(ReportsPrinterView, self).__init__(
            name=name,
            tags_on_print=tags_on_print,
            logger=logger,
        )

        self.report_name = report_name
        self.printer_name = printer_name
        self.pick_list_name = pick_list_name
        self.production_manager = production_manager

        self.nreports = NReports(message_bus=message_bus)
        self.printer_service = PrinterService(
            message_bus=message_bus,
            timeout=timeout,
        )

    def get_printer_name(
        self,  # type: ReportsPrinterView
    ):
        # type: (...) -> str

        return self.printer_name

    def generate_report(
        self,       # type: ReportsPrinterView
        order,      # type: ProductionOrder
    ):
        # type: (...) -> str

        production_order_xml = OrderXml().to_xml(order)
        return self.nreports.request_report(
            pos_id=None,
            report_name=self.report_name,
            production_order=base64.b64encode(production_order_xml).decode("utf-8")
        )

    def print_report(
        self,       # type: ReportsPrinterView
        report,     # type: str
        order,      # type: ProductionOrder
    ):
        # type: (...) -> bool

        for attempt in range(1, MAX_RETRY):
            try:
                printer_logger.info("Sending to printer: {} - Attempt: {}".format(self.printer_name, attempt))

                if attempt == 1:
                    printer_logger.info("Printer: {} - \nPrinting:\n {}".format(self.printer_name, report))

                self.printer_service.send_nreport_to_printer(
                    printer_name=self.printer_name,
                    report_data=report,
                )
                return True
            except (Exception,):
                if attempt >= MAX_RETRY - 1:
                    log_message = "Printer {} unavailable - Failed to Print Report"
                    printer_logger.warning(
                        log_message.format(self.printer_name),
                        exc_info=True,
                    )
                    return False
                else:
                    log_message = "Error when trying to print on Printer {} in attempt {}"
                    printer_logger.warning(
                        log_message.format(self.printer_name),
                        attempt,
                        exc_info=True,
                    )

            time.sleep(2 * attempt)

    def save_report(
        self,               # type: PrinterView
        report,             # type: str
        order,              # type: ProductionOrder
        print_status,       # type: bool
    ):
        # type: (...) -> None

        if self.pick_list_name:
            order.jit_printer_name = self.pick_list_name

        self.production_manager.save_pick_list_order(
            order=order,
            receipt=report,
            status=print_status,
        )

    def clear_printer_cache(
        self,       # type: ReportsPrinterView
    ):
        # type: (...) -> None

        return



import contextlib

from .py39 import import_helper


@contextlib.contextmanager
def isolated_modules():
    """
    Save modules on entry and cleanup on exit.
    """
    (saved,) = import_helper.modules_setup()
    try:
        yield
    finally:
        import_helper.modules_cleanup(saved)


vars(import_helper).setdefault('isolated_modules', isolated_modules)



from abc import (
    ABCMeta,
    abstractmethod,
)

from typing import Dict


class I18n(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def internationalize(
        self,       # type: I18n
        labels,     # type: Dict[str, str]
    ):
        # type: (...) -> Dict[str, str]

        raise NotImplementedError()

    @abstractmethod
    def internationalize_text(
        self,   # type: I18n
        text,   # type: str
    ):
        # type: (...) -> str

        raise NotImplementedError()

    @abstractmethod
    def convert_value_to_currency(
        self,   # type: I18n
        value,  # type: float
    ):
        # type: (...) -> str

        raise NotImplementedError()



from typing import Optional

from sysactions import send_message, check_operator_logged, get_model, action
from msgbus import MBException, TK_CDRAWER_OPEN

from thread_local import get_current_model

@action
def doOpenDrawer(
    pos_id,               # type: str
    check_oper="true",    # type: Optional[str]
    **kwargs
):
    # type: (...) -> None

    model = get_current_model()
    if check_oper == "true":
        check_operator_logged(
            posid=pos_id,
            model=model,
            can_be_blocked=False,
            need_logged=kwargs.get("need_logged", True),
        )

    drawers = model.findall("CashDrawer")
    for drawer in drawers:
        from threading import Thread
        open_drawer_thread = Thread(target=thread_open_drawer, args=(drawer,))
        open_drawer_thread.daemon = True
        open_drawer_thread.start()


def thread_open_drawer(drawer):
    try:
        send_message(drawer.get("name"), TK_CDRAWER_OPEN)
    except MBException:
        pass



import base64
import hashlib
import hmac

from boss.model import User
from boss_api.shared.response import UserDto
from cryptography.fernet import Fernet


def generate_hmac_md5_hash(
    message: str,
    key: str,
) -> str:

    return hmac.new(
        key=key.encode(),
        msg=message.encode(),  # noqa
        digestmod=hashlib.md5,
    ).hexdigest()


def get_encrypt_key(
    user_id: int,
) -> bytes:

    secret = bytes(str(403841032 + int(user_id)), "UTF-8")
    hlib = hashlib.md5()
    hlib.update(secret)
    return base64.urlsafe_b64encode(hlib.hexdigest().encode("UTF-8"))


def decrypt_data(
    user_id: int,
    encrypted_data: str,
) -> str:

    fernet = Fernet(key=get_encrypt_key(user_id=user_id))
    decrypted_data = fernet.decrypt(token=encrypted_data.encode("UTF-8"))
    return decrypted_data.decode("UTF-8")


def encrypt_data(
    user_id: int,
    data: str,
) -> str:

    fernet = Fernet(key=get_encrypt_key(user_id=user_id))
    encrypted_data = fernet.encrypt(str(data).encode("UTF-8"))
    return encrypted_data.decode("UTF-8")


def convert_user_dto_into_user_response(
    user: User,
) -> UserDto:

    return UserDto(
        user_id=user.user_id,
        level=user.level,
        user_name=user.user_name,
        long_name=user.long_name,
        status=user.status,
        admission_date=user.admission_date,
        termination_date=user.termination_date,
        password=user.password,
        fingerprint_minutiae=user.fingerprint_minutiae,
        nfc_uuid=user.nfc_uuid,
    )



# util/_collections.py
# Copyright (C) 2005-2024 the SQLAlchemy authors and contributors
# <see AUTHORS file>
#
# This module is part of SQLAlchemy and is released under
# the MIT License: https://www.opensource.org/licenses/mit-license.php
# mypy: allow-untyped-defs, allow-untyped-calls

"""Collection classes and helpers."""
from __future__ import annotations

import operator
import threading
import types
import typing
from typing import Any
from typing import Callable
from typing import cast
from typing import Container
from typing import Dict
from typing import FrozenSet
from typing import Generic
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Mapping
from typing import NoReturn
from typing import Optional
from typing import overload
from typing import Sequence
from typing import Set
from typing import Tuple
from typing import TypeVar
from typing import Union
from typing import ValuesView
import weakref

from ._has_cy import HAS_CYEXTENSION
from .typing import is_non_string_iterable
from .typing import Literal
from .typing import Protocol

if typing.TYPE_CHECKING or not HAS_CYEXTENSION:
    from ._py_collections import immutabledict as immutabledict
    from ._py_collections import IdentitySet as IdentitySet
    from ._py_collections import ReadOnlyContainer as ReadOnlyContainer
    from ._py_collections import ImmutableDictBase as ImmutableDictBase
    from ._py_collections import OrderedSet as OrderedSet
    from ._py_collections import unique_list as unique_list
else:
    from sqlalchemy.cyextension.immutabledict import (
        ReadOnlyContainer as ReadOnlyContainer,
    )
    from sqlalchemy.cyextension.immutabledict import (
        ImmutableDictBase as ImmutableDictBase,
    )
    from sqlalchemy.cyextension.immutabledict import (
        immutabledict as immutabledict,
    )
    from sqlalchemy.cyextension.collections import IdentitySet as IdentitySet
    from sqlalchemy.cyextension.collections import OrderedSet as OrderedSet
    from sqlalchemy.cyextension.collections import (  # noqa
        unique_list as unique_list,
    )


_T = TypeVar("_T", bound=Any)
_KT = TypeVar("_KT", bound=Any)
_VT = TypeVar("_VT", bound=Any)
_T_co = TypeVar("_T_co", covariant=True)

EMPTY_SET: FrozenSet[Any] = frozenset()
NONE_SET: FrozenSet[Any] = frozenset([None])


def merge_lists_w_ordering(a: List[Any], b: List[Any]) -> List[Any]:
    """merge two lists, maintaining ordering as much as possible.

    this is to reconcile vars(cls) with cls.__annotations__.

    Example::

        >>> a = ['__tablename__', 'id', 'x', 'created_at']
        >>> b = ['id', 'name', 'data', 'y', 'created_at']
        >>> merge_lists_w_ordering(a, b)
        ['__tablename__', 'id', 'name', 'data', 'y', 'x', 'created_at']

    This is not necessarily the ordering that things had on the class,
    in this case the class is::

        class User(Base):
            __tablename__ = "users"

            id: Mapped[int] = mapped_column(primary_key=True)
            name: Mapped[str]
            data: Mapped[Optional[str]]
            x = Column(Integer)
            y: Mapped[int]
            created_at: Mapped[datetime.datetime] = mapped_column()

    But things are *mostly* ordered.

    The algorithm could also be done by creating a partial ordering for
    all items in both lists and then using topological_sort(), but that
    is too much overhead.

    Background on how I came up with this is at:
    https://gist.github.com/zzzeek/89de958cf0803d148e74861bd682ebae

    """
    overlap = set(a).intersection(b)

    result = []

    current, other = iter(a), iter(b)

    while True:
        for element in current:
            if element in overlap:
                overlap.discard(element)
                other, current = current, other
                break

            result.append(element)
        else:
            result.extend(other)
            break

    return result


def coerce_to_immutabledict(d: Mapping[_KT, _VT]) -> immutabledict[_KT, _VT]:
    if not d:
        return EMPTY_DICT
    elif isinstance(d, immutabledict):
        return d
    else:
        return immutabledict(d)


EMPTY_DICT: immutabledict[Any, Any] = immutabledict()


class FacadeDict(ImmutableDictBase[_KT, _VT]):
    """A dictionary that is not publicly mutable."""

    def __new__(cls, *args: Any) -> FacadeDict[Any, Any]:
        new = ImmutableDictBase.__new__(cls)
        return new

    def copy(self) -> NoReturn:
        raise NotImplementedError(
            "an immutabledict shouldn't need to be copied.  use dict(d) "
            "if you need a mutable dictionary."
        )

    def __reduce__(self) -> Any:
        return FacadeDict, (dict(self),)

    def _insert_item(self, key: _KT, value: _VT) -> None:
        """insert an item into the dictionary directly."""
        dict.__setitem__(self, key, value)

    def __repr__(self) -> str:
        return "FacadeDict(%s)" % dict.__repr__(self)


_DT = TypeVar("_DT", bound=Any)

_F = TypeVar("_F", bound=Any)


class Properties(Generic[_T]):
    """Provide a __getattr__/__setattr__ interface over a dict."""

    __slots__ = ("_data",)

    _data: Dict[str, _T]

    def __init__(self, data: Dict[str, _T]):
        object.__setattr__(self, "_data", data)

    def __len__(self) -> int:
        return len(self._data)

    def __iter__(self) -> Iterator[_T]:
        return iter(list(self._data.values()))

    def __dir__(self) -> List[str]:
        return dir(super()) + [str(k) for k in self._data.keys()]

    def __add__(self, other: Properties[_F]) -> List[Union[_T, _F]]:
        return list(self) + list(other)

    def __setitem__(self, key: str, obj: _T) -> None:
        self._data[key] = obj

    def __getitem__(self, key: str) -> _T:
        return self._data[key]

    def __delitem__(self, key: str) -> None:
        del self._data[key]

    def __setattr__(self, key: str, obj: _T) -> None:
        self._data[key] = obj

    def __getstate__(self) -> Dict[str, Any]:
        return {"_data": self._data}

    def __setstate__(self, state: Dict[str, Any]) -> None:
        object.__setattr__(self, "_data", state["_data"])

    def __getattr__(self, key: str) -> _T:
        try:
            return self._data[key]
        except KeyError:
            raise AttributeError(key)

    def __contains__(self, key: str) -> bool:
        return key in self._data

    def as_readonly(self) -> ReadOnlyProperties[_T]:
        """Return an immutable proxy for this :class:`.Properties`."""

        return ReadOnlyProperties(self._data)

    def update(self, value: Dict[str, _T]) -> None:
        self._data.update(value)

    @overload
    def get(self, key: str) -> Optional[_T]: ...

    @overload
    def get(self, key: str, default: Union[_DT, _T]) -> Union[_DT, _T]: ...

    def get(
        self, key: str, default: Optional[Union[_DT, _T]] = None
    ) -> Optional[Union[_T, _DT]]:
        if key in self:
            return self[key]
        else:
            return default

    def keys(self) -> List[str]:
        return list(self._data)

    def values(self) -> List[_T]:
        return list(self._data.values())

    def items(self) -> List[Tuple[str, _T]]:
        return list(self._data.items())

    def has_key(self, key: str) -> bool:
        return key in self._data

    def clear(self) -> None:
        self._data.clear()


class OrderedProperties(Properties[_T]):
    """Provide a __getattr__/__setattr__ interface with an OrderedDict
    as backing store."""

    __slots__ = ()

    def __init__(self):
        Properties.__init__(self, OrderedDict())


class ReadOnlyProperties(ReadOnlyContainer, Properties[_T]):
    """Provide immutable dict/object attribute to an underlying dictionary."""

    __slots__ = ()


def _ordered_dictionary_sort(d, key=None):
    """Sort an OrderedDict in-place."""

    items = [(k, d[k]) for k in sorted(d, key=key)]

    d.clear()

    d.update(items)


OrderedDict = dict
sort_dictionary = _ordered_dictionary_sort


class WeakSequence(Sequence[_T]):
    def __init__(self, __elements: Sequence[_T] = ()):
        # adapted from weakref.WeakKeyDictionary, prevent reference
        # cycles in the collection itself
        def _remove(item, selfref=weakref.ref(self)):
            self = selfref()
            if self is not None:
                self._storage.remove(item)

        self._remove = _remove
        self._storage = [
            weakref.ref(element, _remove) for element in __elements
        ]

    def append(self, item):
        self._storage.append(weakref.ref(item, self._remove))

    def __len__(self):
        return len(self._storage)

    def __iter__(self):
        return (
            obj for obj in (ref() for ref in self._storage) if obj is not None
        )

    def __getitem__(self, index):
        try:
            obj = self._storage[index]
        except KeyError:
            raise IndexError("Index %s out of range" % index)
        else:
            return obj()


class OrderedIdentitySet(IdentitySet):
    def __init__(self, iterable: Optional[Iterable[Any]] = None):
        IdentitySet.__init__(self)
        self._members = OrderedDict()
        if iterable:
            for o in iterable:
                self.add(o)


class PopulateDict(Dict[_KT, _VT]):
    """A dict which populates missing values via a creation function.

    Note the creation function takes a key, unlike
    collections.defaultdict.

    """

    def __init__(self, creator: Callable[[_KT], _VT]):
        self.creator = creator

    def __missing__(self, key: Any) -> Any:
        self[key] = val = self.creator(key)
        return val


class WeakPopulateDict(Dict[_KT, _VT]):
    """Like PopulateDict, but assumes a self + a method and does not create
    a reference cycle.

    """

    def __init__(self, creator_method: types.MethodType):
        self.creator = creator_method.__func__
        weakself = creator_method.__self__
        self.weakself = weakref.ref(weakself)

    def __missing__(self, key: Any) -> Any:
        self[key] = val = self.creator(self.weakself(), key)
        return val


# Define collections that are capable of storing
# ColumnElement objects as hashable keys/elements.
# At this point, these are mostly historical, things
# used to be more complicated.
column_set = set
column_dict = dict
ordered_column_set = OrderedSet


class UniqueAppender(Generic[_T]):
    """Appends items to a collection ensuring uniqueness.

    Additional appends() of the same object are ignored.  Membership is
    determined by identity (``is a``) not equality (``==``).
    """

    __slots__ = "data", "_data_appender", "_unique"

    data: Union[Iterable[_T], Set[_T], List[_T]]
    _data_appender: Callable[[_T], None]
    _unique: Dict[int, Literal[True]]

    def __init__(
        self,
        data: Union[Iterable[_T], Set[_T], List[_T]],
        via: Optional[str] = None,
    ):
        self.data = data
        self._unique = {}
        if via:
            self._data_appender = getattr(data, via)
        elif hasattr(data, "append"):
            self._data_appender = cast("List[_T]", data).append
        elif hasattr(data, "add"):
            self._data_appender = cast("Set[_T]", data).add

    def append(self, item: _T) -> None:
        id_ = id(item)
        if id_ not in self._unique:
            self._data_appender(item)
            self._unique[id_] = True

    def __iter__(self) -> Iterator[_T]:
        return iter(self.data)


def coerce_generator_arg(arg: Any) -> List[Any]:
    if len(arg) == 1 and isinstance(arg[0], types.GeneratorType):
        return list(arg[0])
    else:
        return cast("List[Any]", arg)


def to_list(x: Any, default: Optional[List[Any]] = None) -> List[Any]:
    if x is None:
        return default  # type: ignore
    if not is_non_string_iterable(x):
        return [x]
    elif isinstance(x, list):
        return x
    else:
        return list(x)


def has_intersection(set_: Container[Any], iterable: Iterable[Any]) -> bool:
    r"""return True if any items of set\_ are present in iterable.

    Goes through special effort to ensure __hash__ is not called
    on items in iterable that don't support it.

    """
    return any(i in set_ for i in iterable if i.__hash__)


def to_set(x):
    if x is None:
        return set()
    if not isinstance(x, set):
        return set(to_list(x))
    else:
        return x


def to_column_set(x: Any) -> Set[Any]:
    if x is None:
        return column_set()
    if not isinstance(x, column_set):
        return column_set(to_list(x))
    else:
        return x


def update_copy(d, _new=None, **kw):
    """Copy the given dict and update with the given values."""

    d = d.copy()
    if _new:
        d.update(_new)
    d.update(**kw)
    return d


def flatten_iterator(x: Iterable[_T]) -> Iterator[_T]:
    """Given an iterator of which further sub-elements may also be
    iterators, flatten the sub-elements into a single iterator.

    """
    elem: _T
    for elem in x:
        if not isinstance(elem, str) and hasattr(elem, "__iter__"):
            yield from flatten_iterator(elem)
        else:
            yield elem


class LRUCache(typing.MutableMapping[_KT, _VT]):
    """Dictionary with 'squishy' removal of least
    recently used items.

    Note that either get() or [] should be used here, but
    generally its not safe to do an "in" check first as the dictionary
    can change subsequent to that call.

    """

    __slots__ = (
        "capacity",
        "threshold",
        "size_alert",
        "_data",
        "_counter",
        "_mutex",
    )

    capacity: int
    threshold: float
    size_alert: Optional[Callable[[LRUCache[_KT, _VT]], None]]

    def __init__(
        self,
        capacity: int = 100,
        threshold: float = 0.5,
        size_alert: Optional[Callable[..., None]] = None,
    ):
        self.capacity = capacity
        self.threshold = threshold
        self.size_alert = size_alert
        self._counter = 0
        self._mutex = threading.Lock()
        self._data: Dict[_KT, Tuple[_KT, _VT, List[int]]] = {}

    def _inc_counter(self):
        self._counter += 1
        return self._counter

    @overload
    def get(self, key: _KT) -> Optional[_VT]: ...

    @overload
    def get(self, key: _KT, default: Union[_VT, _T]) -> Union[_VT, _T]: ...

    def get(
        self, key: _KT, default: Optional[Union[_VT, _T]] = None
    ) -> Optional[Union[_VT, _T]]:
        item = self._data.get(key)
        if item is not None:
            item[2][0] = self._inc_counter()
            return item[1]
        else:
            return default

    def __getitem__(self, key: _KT) -> _VT:
        item = self._data[key]
        item[2][0] = self._inc_counter()
        return item[1]

    def __iter__(self) -> Iterator[_KT]:
        return iter(self._data)

    def __len__(self) -> int:
        return len(self._data)

    def values(self) -> ValuesView[_VT]:
        return typing.ValuesView({k: i[1] for k, i in self._data.items()})

    def __setitem__(self, key: _KT, value: _VT) -> None:
        self._data[key] = (key, value, [self._inc_counter()])
        self._manage_size()

    def __delitem__(self, __v: _KT) -> None:
        del self._data[__v]

    @property
    def size_threshold(self) -> float:
        return self.capacity + self.capacity * self.threshold

    def _manage_size(self) -> None:
        if not self._mutex.acquire(False):
            return
        try:
            size_alert = bool(self.size_alert)
            while len(self) > self.capacity + self.capacity * self.threshold:
                if size_alert:
                    size_alert = False
                    self.size_alert(self)  # type: ignore
                by_counter = sorted(
                    self._data.values(),
                    key=operator.itemgetter(2),
                    reverse=True,
                )
                for item in by_counter[self.capacity :]:
                    try:
                        del self._data[item[0]]
                    except KeyError:
                        # deleted elsewhere; skip
                        continue
        finally:
            self._mutex.release()


class _CreateFuncType(Protocol[_T_co]):
    def __call__(self) -> _T_co: ...


class _ScopeFuncType(Protocol):
    def __call__(self) -> Any: ...


class ScopedRegistry(Generic[_T]):
    """A Registry that can store one or multiple instances of a single
    class on the basis of a "scope" function.

    The object implements ``__call__`` as the "getter", so by
    calling ``myregistry()`` the contained object is returned
    for the current scope.

    :param createfunc:
      a callable that returns a new object to be placed in the registry

    :param scopefunc:
      a callable that will return a key to store/retrieve an object.
    """

    __slots__ = "createfunc", "scopefunc", "registry"

    createfunc: _CreateFuncType[_T]
    scopefunc: _ScopeFuncType
    registry: Any

    def __init__(
        self, createfunc: Callable[[], _T], scopefunc: Callable[[], Any]
    ):
        """Construct a new :class:`.ScopedRegistry`.

        :param createfunc:  A creation function that will generate
          a new value for the current scope, if none is present.

        :param scopefunc:  A function that returns a hashable
          token representing the current scope (such as, current
          thread identifier).

        """
        self.createfunc = createfunc
        self.scopefunc = scopefunc
        self.registry = {}

    def __call__(self) -> _T:
        key = self.scopefunc()
        try:
            return self.registry[key]  # type: ignore[no-any-return]
        except KeyError:
            return self.registry.setdefault(key, self.createfunc())  # type: ignore[no-any-return] # noqa: E501

    def has(self) -> bool:
        """Return True if an object is present in the current scope."""

        return self.scopefunc() in self.registry

    def set(self, obj: _T) -> None:
        """Set the value for the current scope."""

        self.registry[self.scopefunc()] = obj

    def clear(self) -> None:
        """Clear the current scope, if any."""

        try:
            del self.registry[self.scopefunc()]
        except KeyError:
            pass


class ThreadLocalRegistry(ScopedRegistry[_T]):
    """A :class:`.ScopedRegistry` that uses a ``threading.local()``
    variable for storage.

    """

    def __init__(self, createfunc: Callable[[], _T]):
        self.createfunc = createfunc
        self.registry = threading.local()

    def __call__(self) -> _T:
        try:
            return self.registry.value  # type: ignore[no-any-return]
        except AttributeError:
            val = self.registry.value = self.createfunc()
            return val

    def has(self) -> bool:
        return hasattr(self.registry, "value")

    def set(self, obj: _T) -> None:
        self.registry.value = obj

    def clear(self) -> None:
        try:
            del self.registry.value
        except AttributeError:
            pass


def has_dupes(sequence, target):
    """Given a sequence and search object, return True if there's more
    than one, False if zero or one of them.


    """
    # compare to .index version below, this version introduces less function
    # overhead and is usually the same speed.  At 15000 items (way bigger than
    # a relationship-bound collection in memory usually is) it begins to
    # fall behind the other version only by microseconds.
    c = 0
    for item in sequence:
        if item is target:
            c += 1
            if c > 1:
                return True
    return False


# .index version.  the two __contains__ calls as well
# as .index() and isinstance() slow this down.
# def has_dupes(sequence, target):
#    if target not in sequence:
#        return False
#    elif not isinstance(sequence, collections_abc.Sequence):
#        return False
#
#    idx = sequence.index(target)
#    return target in sequence[idx + 1:]



from ._CurrentTabWithPartialTenders import CurrentTabWithPartialTenders
from ._InvalidTableStatus import InvalidTableStatus
from ._InvalidTabMask import InvalidTabMask
from ._InvalidTabName import InvalidTabName
from ._MissingAnyOfParameters import MissingAnyOfParameters
from ._MissingNumSeats import MissingNumSeats
from ._OperatorNotLoggedIn import OperatorNotLoggedIn
from ._TabIsNotASplitSeat import TabIsNotASplitSeat
from ._TabNameAlreadyExists import TabNameAlreadyExists
from ._TabNotFound import TabNotFound
from ._TabToJoinDoesNotExist import TabToJoinDoesNotExists
from ._TabToJoinIsEmpty import TabToJoinIsEmpty
from ._TabToJoinInInvalidStatus import TabToJoinInInvalidStatus
from ._TabToJoinIsAlreadyJoined import TabToJoinIsAlreadyJoined
from ._TabToJoinIsTheSame import TabToJoinIsTheSame
from ._TabToJoinOpenedInAnotherPOS import TabToJoinOpenedInAnotherPOS
from ._TabToJoinWithPartialTenders import TabToJoinWithPartialTenders
from ._TableAlreadyOpened import TableAlreadyOpened
from ._TableError import TableError
from ._TableException import TableException
from ._TableIsNotInTabsMode import TableIsNotInTabsMode
from ._TableNotFound import TableNotFound
from ._TabNotInValidTabNames import TabNotInValidTabNames
from ._TableNotOpened import TableNotOpened
from ._TableServiceException import TableServiceException
from ._UnknownError import UnknownError



from typing import (
    Dict,
    List,
    Union,
    NewType,
)

from ._HTMLTextFormatter import HTMLTextFormatter

JSON = Union[Dict[str, "JSON"], List["JSON"], str, int, float, bool, None]  # type: NewType("JSON", JSON)


class LineParser(object):

    def __init__(
        self,       # type: LineParser
        columns,    # type: int
    ):
        # type: (...) -> None

        self.columns = columns

    def parse(
        self,    # type: LineParser
        line,    # type: JSON
    ):
        # type: (...) -> List[str]

        lines = self._get_lines_split_by_columns(text=line.get("text"))
        return self._format_lines(
            lines=lines,
            customization=line.get("customization"),
        )

    def _format_lines(
        self,             # type: LineParser
        lines,            # type: List[str]
        customization,    # type: Dict[str, Union[str, bool, int, float]]
    ):
        # type: (...) -> List[str]

        formatted = []
        html_text_formatter = HTMLTextFormatter(
            customization=customization,
            columns=self.columns,
        )
        for line in lines:
            formatted.append(html_text_formatter.format(text=line))
            formatted.append("<br/>")

        formatted.pop()

        return formatted

    def _get_lines_split_by_columns(
        self,    # type: LineParser
        text,    # type: str
    ):
        # type: (...) -> List[str]

        if len(text) <= self.columns:
            return [text]
        elif text.count("\\n") > 0:
            return text.split("\\n")

        lines = []
        words = text.split()
        current_line = ""
        index = 0
        while True:
            current_line += words[index] + " "
            index += 1
            if index >= len(words):
                lines.append(current_line.strip())
                break

            if len(current_line + words[index]) > self.columns:
                lines.append(current_line.strip())
                current_line = ""

        return lines



# orm/writeonly.py
# Copyright (C) 2005-2024 the SQLAlchemy authors and contributors
# <see AUTHORS file>
#
# This module is part of SQLAlchemy and is released under
# the MIT License: https://www.opensource.org/licenses/mit-license.php

"""Write-only collection API.

This is an alternate mapped attribute style that only supports single-item
collection mutation operations.   To read the collection, a select()
object must be executed each time.

.. versionadded:: 2.0


"""

from __future__ import annotations

from typing import Any
from typing import Collection
from typing import Dict
from typing import Generic
from typing import Iterable
from typing import Iterator
from typing import List
from typing import NoReturn
from typing import Optional
from typing import overload
from typing import Tuple
from typing import Type
from typing import TYPE_CHECKING
from typing import TypeVar
from typing import Union

from sqlalchemy.sql import bindparam
from . import attributes
from . import interfaces
from . import relationships
from . import strategies
from .base import NEVER_SET
from .base import object_mapper
from .base import PassiveFlag
from .base import RelationshipDirection
from .. import exc
from .. import inspect
from .. import log
from .. import util
from ..sql import delete
from ..sql import insert
from ..sql import select
from ..sql import update
from ..sql.dml import Delete
from ..sql.dml import Insert
from ..sql.dml import Update
from ..util.typing import Literal

if TYPE_CHECKING:
    from . import QueryableAttribute
    from ._typing import _InstanceDict
    from .attributes import AttributeEventToken
    from .base import LoaderCallableStatus
    from .collections import _AdaptedCollectionProtocol
    from .collections import CollectionAdapter
    from .mapper import Mapper
    from .relationships import _RelationshipOrderByArg
    from .state import InstanceState
    from .util import AliasedClass
    from ..event import _Dispatch
    from ..sql.selectable import FromClause
    from ..sql.selectable import Select

_T = TypeVar("_T", bound=Any)


class WriteOnlyHistory(Generic[_T]):
    """Overrides AttributeHistory to receive append/remove events directly."""

    unchanged_items: util.OrderedIdentitySet
    added_items: util.OrderedIdentitySet
    deleted_items: util.OrderedIdentitySet
    _reconcile_collection: bool

    def __init__(
        self,
        attr: WriteOnlyAttributeImpl,
        state: InstanceState[_T],
        passive: PassiveFlag,
        apply_to: Optional[WriteOnlyHistory[_T]] = None,
    ) -> None:
        if apply_to:
            if passive & PassiveFlag.SQL_OK:
                raise exc.InvalidRequestError(
                    f"Attribute {attr} can't load the existing state from the "
                    "database for this operation; full iteration is not "
                    "permitted.  If this is a delete operation, configure "
                    f"passive_deletes=True on the {attr} relationship in "
                    "order to resolve this error."
                )

            self.unchanged_items = apply_to.unchanged_items
            self.added_items = apply_to.added_items
            self.deleted_items = apply_to.deleted_items
            self._reconcile_collection = apply_to._reconcile_collection
        else:
            self.deleted_items = util.OrderedIdentitySet()
            self.added_items = util.OrderedIdentitySet()
            self.unchanged_items = util.OrderedIdentitySet()
            self._reconcile_collection = False

    @property
    def added_plus_unchanged(self) -> List[_T]:
        return list(self.added_items.union(self.unchanged_items))

    @property
    def all_items(self) -> List[_T]:
        return list(
            self.added_items.union(self.unchanged_items).union(
                self.deleted_items
            )
        )

    def as_history(self) -> attributes.History:
        if self._reconcile_collection:
            added = self.added_items.difference(self.unchanged_items)
            deleted = self.deleted_items.intersection(self.unchanged_items)
            unchanged = self.unchanged_items.difference(deleted)
        else:
            added, unchanged, deleted = (
                self.added_items,
                self.unchanged_items,
                self.deleted_items,
            )
        return attributes.History(list(added), list(unchanged), list(deleted))

    def indexed(self, index: Union[int, slice]) -> Union[List[_T], _T]:
        return list(self.added_items)[index]

    def add_added(self, value: _T) -> None:
        self.added_items.add(value)

    def add_removed(self, value: _T) -> None:
        if value in self.added_items:
            self.added_items.remove(value)
        else:
            self.deleted_items.add(value)


class WriteOnlyAttributeImpl(
    attributes.HasCollectionAdapter, attributes.AttributeImpl
):
    uses_objects: bool = True
    default_accepts_scalar_loader: bool = False
    supports_population: bool = False
    _supports_dynamic_iteration: bool = False
    collection: bool = False
    dynamic: bool = True
    order_by: _RelationshipOrderByArg = ()
    collection_history_cls: Type[WriteOnlyHistory[Any]] = WriteOnlyHistory

    query_class: Type[WriteOnlyCollection[Any]]

    def __init__(
        self,
        class_: Union[Type[Any], AliasedClass[Any]],
        key: str,
        dispatch: _Dispatch[QueryableAttribute[Any]],
        target_mapper: Mapper[_T],
        order_by: _RelationshipOrderByArg,
        **kw: Any,
    ):
        super().__init__(class_, key, None, dispatch, **kw)
        self.target_mapper = target_mapper
        self.query_class = WriteOnlyCollection
        if order_by:
            self.order_by = tuple(order_by)

    def get(
        self,
        state: InstanceState[Any],
        dict_: _InstanceDict,
        passive: PassiveFlag = PassiveFlag.PASSIVE_OFF,
    ) -> Union[util.OrderedIdentitySet, WriteOnlyCollection[Any]]:
        if not passive & PassiveFlag.SQL_OK:
            return self._get_collection_history(
                state, PassiveFlag.PASSIVE_NO_INITIALIZE
            ).added_items
        else:
            return self.query_class(self, state)

    @overload
    def get_collection(
        self,
        state: InstanceState[Any],
        dict_: _InstanceDict,
        user_data: Literal[None] = ...,
        passive: Literal[PassiveFlag.PASSIVE_OFF] = ...,
    ) -> CollectionAdapter: ...

    @overload
    def get_collection(
        self,
        state: InstanceState[Any],
        dict_: _InstanceDict,
        user_data: _AdaptedCollectionProtocol = ...,
        passive: PassiveFlag = ...,
    ) -> CollectionAdapter: ...

    @overload
    def get_collection(
        self,
        state: InstanceState[Any],
        dict_: _InstanceDict,
        user_data: Optional[_AdaptedCollectionProtocol] = ...,
        passive: PassiveFlag = ...,
    ) -> Union[
        Literal[LoaderCallableStatus.PASSIVE_NO_RESULT], CollectionAdapter
    ]: ...

    def get_collection(
        self,
        state: InstanceState[Any],
        dict_: _InstanceDict,
        user_data: Optional[_AdaptedCollectionProtocol] = None,
        passive: PassiveFlag = PassiveFlag.PASSIVE_OFF,
    ) -> Union[
        Literal[LoaderCallableStatus.PASSIVE_NO_RESULT], CollectionAdapter
    ]:
        data: Collection[Any]
        if not passive & PassiveFlag.SQL_OK:
            data = self._get_collection_history(state, passive).added_items
        else:
            history = self._get_collection_history(state, passive)
            data = history.added_plus_unchanged
        return DynamicCollectionAdapter(data)  # type: ignore[return-value]

    @util.memoized_property
    def _append_token(  # type:ignore[override]
        self,
    ) -> attributes.AttributeEventToken:
        return attributes.AttributeEventToken(self, attributes.OP_APPEND)

    @util.memoized_property
    def _remove_token(  # type:ignore[override]
        self,
    ) -> attributes.AttributeEventToken:
        return attributes.AttributeEventToken(self, attributes.OP_REMOVE)

    def fire_append_event(
        self,
        state: InstanceState[Any],
        dict_: _InstanceDict,
        value: Any,
        initiator: Optional[AttributeEventToken],
        collection_history: Optional[WriteOnlyHistory[Any]] = None,
    ) -> None:
        if collection_history is None:
            collection_history = self._modified_event(state, dict_)

        collection_history.add_added(value)

        for fn in self.dispatch.append:
            value = fn(state, value, initiator or self._append_token)

        if self.trackparent and value is not None:
            self.sethasparent(attributes.instance_state(value), state, True)

    def fire_remove_event(
        self,
        state: InstanceState[Any],
        dict_: _InstanceDict,
        value: Any,
        initiator: Optional[AttributeEventToken],
        collection_history: Optional[WriteOnlyHistory[Any]] = None,
    ) -> None:
        if collection_history is None:
            collection_history = self._modified_event(state, dict_)

        collection_history.add_removed(value)

        if self.trackparent and value is not None:
            self.sethasparent(attributes.instance_state(value), state, False)

        for fn in self.dispatch.remove:
            fn(state, value, initiator or self._remove_token)

    def _modified_event(
        self, state: InstanceState[Any], dict_: _InstanceDict
    ) -> WriteOnlyHistory[Any]:
        if self.key not in state.committed_state:
            state.committed_state[self.key] = self.collection_history_cls(
                self, state, PassiveFlag.PASSIVE_NO_FETCH
            )

        state._modified_event(dict_, self, NEVER_SET)

        # this is a hack to allow the entities.ComparableEntity fixture
        # to work
        dict_[self.key] = True
        return state.committed_state[self.key]  # type: ignore[no-any-return]

    def set(
        self,
        state: InstanceState[Any],
        dict_: _InstanceDict,
        value: Any,
        initiator: Optional[AttributeEventToken] = None,
        passive: PassiveFlag = PassiveFlag.PASSIVE_OFF,
        check_old: Any = None,
        pop: bool = False,
        _adapt: bool = True,
    ) -> None:
        if initiator and initiator.parent_token is self.parent_token:
            return

        if pop and value is None:
            return

        iterable = value
        new_values = list(iterable)
        if state.has_identity:
            if not self._supports_dynamic_iteration:
                raise exc.InvalidRequestError(
                    f'Collection "{self}" does not support implicit '
                    "iteration; collection replacement operations "
                    "can't be used"
                )
            old_collection = util.IdentitySet(
                self.get(state, dict_, passive=passive)
            )

        collection_history = self._modified_event(state, dict_)
        if not state.has_identity:
            old_collection = collection_history.added_items
        else:
            old_collection = old_collection.union(
                collection_history.added_items
            )

        constants = old_collection.intersection(new_values)
        additions = util.IdentitySet(new_values).difference(constants)
        removals = old_collection.difference(constants)

        for member in new_values:
            if member in additions:
                self.fire_append_event(
                    state,
                    dict_,
                    member,
                    None,
                    collection_history=collection_history,
                )

        for member in removals:
            self.fire_remove_event(
                state,
                dict_,
                member,
                None,
                collection_history=collection_history,
            )

    def delete(self, *args: Any, **kwargs: Any) -> NoReturn:
        raise NotImplementedError()

    def set_committed_value(
        self, state: InstanceState[Any], dict_: _InstanceDict, value: Any
    ) -> NoReturn:
        raise NotImplementedError(
            "Dynamic attributes don't support collection population."
        )

    def get_history(
        self,
        state: InstanceState[Any],
        dict_: _InstanceDict,
        passive: PassiveFlag = PassiveFlag.PASSIVE_NO_FETCH,
    ) -> attributes.History:
        c = self._get_collection_history(state, passive)
        return c.as_history()

    def get_all_pending(
        self,
        state: InstanceState[Any],
        dict_: _InstanceDict,
        passive: PassiveFlag = PassiveFlag.PASSIVE_NO_INITIALIZE,
    ) -> List[Tuple[InstanceState[Any], Any]]:
        c = self._get_collection_history(state, passive)
        return [(attributes.instance_state(x), x) for x in c.all_items]

    def _get_collection_history(
        self, state: InstanceState[Any], passive: PassiveFlag
    ) -> WriteOnlyHistory[Any]:
        c: WriteOnlyHistory[Any]
        if self.key in state.committed_state:
            c = state.committed_state[self.key]
        else:
            c = self.collection_history_cls(
                self, state, PassiveFlag.PASSIVE_NO_FETCH
            )

        if state.has_identity and (passive & PassiveFlag.INIT_OK):
            return self.collection_history_cls(
                self, state, passive, apply_to=c
            )
        else:
            return c

    def append(
        self,
        state: InstanceState[Any],
        dict_: _InstanceDict,
        value: Any,
        initiator: Optional[AttributeEventToken],
        passive: PassiveFlag = PassiveFlag.PASSIVE_NO_FETCH,
    ) -> None:
        if initiator is not self:
            self.fire_append_event(state, dict_, value, initiator)

    def remove(
        self,
        state: InstanceState[Any],
        dict_: _InstanceDict,
        value: Any,
        initiator: Optional[AttributeEventToken],
        passive: PassiveFlag = PassiveFlag.PASSIVE_NO_FETCH,
    ) -> None:
        if initiator is not self:
            self.fire_remove_event(state, dict_, value, initiator)

    def pop(
        self,
        state: InstanceState[Any],
        dict_: _InstanceDict,
        value: Any,
        initiator: Optional[AttributeEventToken],
        passive: PassiveFlag = PassiveFlag.PASSIVE_NO_FETCH,
    ) -> None:
        self.remove(state, dict_, value, initiator, passive=passive)


@log.class_logger
@relationships.RelationshipProperty.strategy_for(lazy="write_only")
class WriteOnlyLoader(strategies.AbstractRelationshipLoader, log.Identified):
    impl_class = WriteOnlyAttributeImpl

    def init_class_attribute(self, mapper: Mapper[Any]) -> None:
        self.is_class_level = True
        if not self.uselist or self.parent_property.direction not in (
            interfaces.ONETOMANY,
            interfaces.MANYTOMANY,
        ):
            raise exc.InvalidRequestError(
                "On relationship %s, 'dynamic' loaders cannot be used with "
                "many-to-one/one-to-one relationships and/or "
                "uselist=False." % self.parent_property
            )

        strategies._register_attribute(  # type: ignore[no-untyped-call]
            self.parent_property,
            mapper,
            useobject=True,
            impl_class=self.impl_class,
            target_mapper=self.parent_property.mapper,
            order_by=self.parent_property.order_by,
            query_class=self.parent_property.query_class,
        )


class DynamicCollectionAdapter:
    """simplified CollectionAdapter for internal API consistency"""

    data: Collection[Any]

    def __init__(self, data: Collection[Any]):
        self.data = data

    def __iter__(self) -> Iterator[Any]:
        return iter(self.data)

    def _reset_empty(self) -> None:
        pass

    def __len__(self) -> int:
        return len(self.data)

    def __bool__(self) -> bool:
        return True


class AbstractCollectionWriter(Generic[_T]):
    """Virtual collection which includes append/remove methods that synchronize
    into the attribute event system.

    """

    if not TYPE_CHECKING:
        __slots__ = ()

    instance: _T
    _from_obj: Tuple[FromClause, ...]

    def __init__(self, attr: WriteOnlyAttributeImpl, state: InstanceState[_T]):
        instance = state.obj()
        if TYPE_CHECKING:
            assert instance
        self.instance = instance
        self.attr = attr

        mapper = object_mapper(instance)
        prop = mapper._props[self.attr.key]

        if prop.secondary is not None:
            # this is a hack right now.  The Query only knows how to
            # make subsequent joins() without a given left-hand side
            # from self._from_obj[0].  We need to ensure prop.secondary
            # is in the FROM.  So we purposely put the mapper selectable
            # in _from_obj[0] to ensure a user-defined join() later on
            # doesn't fail, and secondary is then in _from_obj[1].

            # note also, we are using the official ORM-annotated selectable
            # from __clause_element__(), see #7868
            self._from_obj = (prop.mapper.__clause_element__(), prop.secondary)
        else:
            self._from_obj = ()

        self._where_criteria = (
            prop._with_parent(instance, alias_secondary=False),
        )

        if self.attr.order_by:
            self._order_by_clauses = self.attr.order_by
        else:
            self._order_by_clauses = ()

    def _add_all_impl(self, iterator: Iterable[_T]) -> None:
        for item in iterator:
            self.attr.append(
                attributes.instance_state(self.instance),
                attributes.instance_dict(self.instance),
                item,
                None,
            )

    def _remove_impl(self, item: _T) -> None:
        self.attr.remove(
            attributes.instance_state(self.instance),
            attributes.instance_dict(self.instance),
            item,
            None,
        )


class WriteOnlyCollection(AbstractCollectionWriter[_T]):
    """Write-only collection which can synchronize changes into the
    attribute event system.

    The :class:`.WriteOnlyCollection` is used in a mapping by
    using the ``"write_only"`` lazy loading strategy with
    :func:`_orm.relationship`.     For background on this configuration,
    see :ref:`write_only_relationship`.

    .. versionadded:: 2.0

    .. seealso::

        :ref:`write_only_relationship`

    """

    __slots__ = (
        "instance",
        "attr",
        "_where_criteria",
        "_from_obj",
        "_order_by_clauses",
    )

    def __iter__(self) -> NoReturn:
        raise TypeError(
            "WriteOnly collections don't support iteration in-place; "
            "to query for collection items, use the select() method to "
            "produce a SQL statement and execute it with session.scalars()."
        )

    def select(self) -> Select[Tuple[_T]]:
        """Produce a :class:`_sql.Select` construct that represents the
        rows within this instance-local :class:`_orm.WriteOnlyCollection`.

        """
        stmt = select(self.attr.target_mapper).where(*self._where_criteria)
        if self._from_obj:
            stmt = stmt.select_from(*self._from_obj)
        if self._order_by_clauses:
            stmt = stmt.order_by(*self._order_by_clauses)
        return stmt

    def insert(self) -> Insert:
        """For one-to-many collections, produce a :class:`_dml.Insert` which
        will insert new rows in terms of this this instance-local
        :class:`_orm.WriteOnlyCollection`.

        This construct is only supported for a :class:`_orm.Relationship`
        that does **not** include the :paramref:`_orm.relationship.secondary`
        parameter.  For relationships that refer to a many-to-many table,
        use ordinary bulk insert techniques to produce new objects, then
        use :meth:`_orm.AbstractCollectionWriter.add_all` to associate them
        with the collection.


        """

        state = inspect(self.instance)
        mapper = state.mapper
        prop = mapper._props[self.attr.key]

        if prop.direction is not RelationshipDirection.ONETOMANY:
            raise exc.InvalidRequestError(
                "Write only bulk INSERT only supported for one-to-many "
                "collections; for many-to-many, use a separate bulk "
                "INSERT along with add_all()."
            )

        dict_: Dict[str, Any] = {}

        for l, r in prop.synchronize_pairs:
            fn = prop._get_attr_w_warn_on_none(
                mapper,
                state,
                state.dict,
                l,
            )

            dict_[r.key] = bindparam(None, callable_=fn)

        return insert(self.attr.target_mapper).values(**dict_)

    def update(self) -> Update:
        """Produce a :class:`_dml.Update` which will refer to rows in terms
        of this instance-local :class:`_orm.WriteOnlyCollection`.

        """
        return update(self.attr.target_mapper).where(*self._where_criteria)

    def delete(self) -> Delete:
        """Produce a :class:`_dml.Delete` which will refer to rows in terms
        of this instance-local :class:`_orm.WriteOnlyCollection`.

        """
        return delete(self.attr.target_mapper).where(*self._where_criteria)

    def add_all(self, iterator: Iterable[_T]) -> None:
        """Add an iterable of items to this :class:`_orm.WriteOnlyCollection`.

        The given items will be persisted to the database in terms of
        the parent instance's collection on the next flush.

        """
        self._add_all_impl(iterator)

    def add(self, item: _T) -> None:
        """Add an item to this :class:`_orm.WriteOnlyCollection`.

        The given item will be persisted to the database in terms of
        the parent instance's collection on the next flush.

        """
        self._add_all_impl([item])

    def remove(self, item: _T) -> None:
        """Remove an item from this :class:`_orm.WriteOnlyCollection`.

        The given item will be removed from the parent instance's collection on
        the next flush.

        """
        self._remove_impl(item)



from ._UuidMessageProcessor import UuidMessageProcessor
from ._LoggingProcessorCallback import (
    LoggingProcessorCallback,
    LogConfig,
    CallBackMethod,
)
from ._TimeProcessorCallback import TimeProcessorCallback



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import Final

OTEL_SCOPE_NAME: Final = "otel.scope.name"
"""
The name of the instrumentation scope - (`InstrumentationScope.Name` in OTLP).
"""

OTEL_SCOPE_VERSION: Final = "otel.scope.version"
"""
The version of the instrumentation scope - (`InstrumentationScope.Version` in OTLP).
"""

OTEL_STATUS_CODE: Final = "otel.status_code"
"""
Name of the code, either "OK" or "ERROR". MUST NOT be set if the status code is UNSET.
"""

OTEL_STATUS_DESCRIPTION: Final = "otel.status_description"
"""
Description of the Status if it has a value, otherwise not set.
"""


class OtelStatusCodeValues(Enum):
    OK = "OK"
    """The operation has been validated by an Application developer or Operator to have completed successfully."""
    ERROR = "ERROR"
    """The operation contains an error."""



class Table(object):

    def __init__(
        self,                       # type: Table
        id="",                      # type: str
        name="",                    # type: str
        current_pos_id="",          # type: str
        status="",                  # type: str
        business_period="",         # type: str
        sector="",                  # type: str
        slice_parent_table_id="",   # type: str
        slice_source_table_id="",   # type: str
        mode="",                    # type: str
        sub_sector="",              # type: str
        user_id="",                 # type: str
        sub_sector_name="",         # type: str
    ):
        # type: (...) -> None

        self.id = id
        self.name = name
        self.current_pos_id = current_pos_id
        self.status = int(status) if status != "" else None
        self.business_period = business_period
        self.sector = sector
        self.slice_source_table_id = slice_source_table_id
        self.slice_parent_table_id = slice_parent_table_id
        self.mode = mode
        self.sub_sector = sub_sector
        self.user_id = user_id
        self.sub_sector_name = sub_sector_name



import unicodedata
from hmac import compare_digest
from typing import Dict, Optional, Union
from urllib.parse import quote, urlencode, urlparse


def build_uri(
    secret: str,
    name: str,
    initial_count: Optional[int] = None,
    issuer: Optional[str] = None,
    algorithm: Optional[str] = None,
    digits: Optional[int] = None,
    period: Optional[int] = None,
    image: Optional[str] = None,
) -> str:
    """
    Returns the provisioning URI for the OTP; works for either TOTP or HOTP.

    This can then be encoded in a QR Code and used to provision the Google
    Authenticator app.

    For module-internal use.

    See also:
        https://github.com/google/google-authenticator/wiki/Key-Uri-Format

    :param secret: the hotp/totp secret used to generate the URI
    :param name: name of the account
    :param initial_count: starting counter value, defaults to None.
        If none, the OTP type will be assumed as TOTP.
    :param issuer: the name of the OTP issuer; this will be the
        organization title of the OTP entry in Authenticator
    :param algorithm: the algorithm used in the OTP generation.
    :param digits: the length of the OTP generated code.
    :param period: the number of seconds the OTP generator is set to
        expire every code.
    :param image: optional logo image url
    :returns: provisioning uri
    """
    # initial_count may be 0 as a valid param
    is_initial_count_present = initial_count is not None

    # Handling values different from defaults
    is_algorithm_set = algorithm is not None and algorithm != "sha1"
    is_digits_set = digits is not None and digits != 6
    is_period_set = period is not None and period != 30

    otp_type = "hotp" if is_initial_count_present else "totp"
    base_uri = "otpauth://{0}/{1}?{2}"

    url_args: Dict[str, Union[None, int, str]] = {"secret": secret}

    label = quote(name)
    if issuer is not None:
        label = quote(issuer) + ":" + label
        url_args["issuer"] = issuer

    if is_initial_count_present:
        url_args["counter"] = initial_count
    if is_algorithm_set:
        url_args["algorithm"] = algorithm.upper()  # type: ignore
    if is_digits_set:
        url_args["digits"] = digits
    if is_period_set:
        url_args["period"] = period
    if image:
        image_uri = urlparse(image)
        if image_uri.scheme != "https" or not image_uri.netloc or not image_uri.path:
            raise ValueError("{} is not a valid url".format(image_uri))
        url_args["image"] = image

    uri = base_uri.format(otp_type, label, urlencode(url_args).replace("+", "%20"))
    return uri


def strings_equal(s1: str, s2: str) -> bool:
    """
    Timing-attack resistant string comparison.

    Normal comparison using == will short-circuit on the first mismatching
    character. This avoids that by scanning the whole string, though we
    still reveal to a timing attack whether the strings are the same
    length.
    """
    s1 = unicodedata.normalize("NFKC", s1)
    s2 = unicodedata.normalize("NFKC", s2)
    return compare_digest(s1.encode("utf-8"), s2.encode("utf-8"))



from ._TableError import TableError


class TabToJoinIsTheSame(TableError):

    def __init__(
        self,       # type: TabToJoinIsTheSame
    ):
        # type: (...) -> None
        super(TabToJoinIsTheSame, self).__init__("$SELECT_A_DIFFERENT_TAB")

    def get_extra_data(
        self,   # type: TabToJoinIsTheSame
    ):
        # type: (...) -> None

        return None



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import (
    Callable,
    Final,
    Generator,
    Iterable,
    Optional,
    Sequence,
    Union,
)

from opentelemetry.metrics import (
    CallbackOptions,
    Counter,
    Meter,
    ObservableGauge,
    Observation,
    UpDownCounter,
)

# pylint: disable=invalid-name
CallbackT = Union[
    Callable[[CallbackOptions], Iterable[Observation]],
    Generator[Iterable[Observation], CallbackOptions, None],
]

HW_ENERGY: Final = "hw.energy"
"""
Energy consumed by the component
Instrument: counter
Unit: J
"""


def create_hw_energy(meter: Meter) -> Counter:
    """Energy consumed by the component"""
    return meter.create_counter(
        name=HW_ENERGY,
        description="Energy consumed by the component",
        unit="J",
    )


HW_ERRORS: Final = "hw.errors"
"""
Number of errors encountered by the component
Instrument: counter
Unit: {error}
"""


def create_hw_errors(meter: Meter) -> Counter:
    """Number of errors encountered by the component"""
    return meter.create_counter(
        name=HW_ERRORS,
        description="Number of errors encountered by the component",
        unit="{error}",
    )


HW_POWER: Final = "hw.power"
"""
Instantaneous power consumed by the component
Instrument: gauge
Unit: W
Note: It is recommended to report `hw.energy` instead of `hw.power` when possible.
"""


def create_hw_power(
    meter: Meter, callbacks: Optional[Sequence[CallbackT]]
) -> ObservableGauge:
    """Instantaneous power consumed by the component"""
    return meter.create_observable_gauge(
        name=HW_POWER,
        callbacks=callbacks,
        description="Instantaneous power consumed by the component",
        unit="W",
    )


HW_STATUS: Final = "hw.status"
"""
Operational status: `1` (true) or `0` (false) for each of the possible states
Instrument: updowncounter
Unit: 1
Note: `hw.status` is currently specified as an *UpDownCounter* but would ideally be represented using a [*StateSet* as defined in OpenMetrics](https://github.com/OpenObservability/OpenMetrics/blob/main/specification/OpenMetrics.md#stateset). This semantic convention will be updated once *StateSet* is specified in OpenTelemetry. This planned change is not expected to have any consequence on the way users query their timeseries backend to retrieve the values of `hw.status` over time.
"""


def create_hw_status(meter: Meter) -> UpDownCounter:
    """Operational status: `1` (true) or `0` (false) for each of the possible states"""
    return meter.create_up_down_counter(
        name=HW_STATUS,
        description="Operational status: `1` (true) or `0` (false) for each of the possible states",
        unit="1",
    )



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.

from __future__ import annotations

import typing

from cryptography.hazmat.primitives.ciphers import Cipher
from cryptography.hazmat.primitives.ciphers.algorithms import AES
from cryptography.hazmat.primitives.ciphers.modes import ECB
from cryptography.hazmat.primitives.constant_time import bytes_eq


def _wrap_core(
    wrapping_key: bytes,
    a: bytes,
    r: list[bytes],
) -> bytes:
    # RFC 3394 Key Wrap - 2.2.1 (index method)
    encryptor = Cipher(AES(wrapping_key), ECB()).encryptor()
    n = len(r)
    for j in range(6):
        for i in range(n):
            # every encryption operation is a discrete 16 byte chunk (because
            # AES has a 128-bit block size) and since we're using ECB it is
            # safe to reuse the encryptor for the entire operation
            b = encryptor.update(a + r[i])
            a = (
                int.from_bytes(b[:8], byteorder="big") ^ ((n * j) + i + 1)
            ).to_bytes(length=8, byteorder="big")
            r[i] = b[-8:]

    assert encryptor.finalize() == b""

    return a + b"".join(r)


def aes_key_wrap(
    wrapping_key: bytes,
    key_to_wrap: bytes,
    backend: typing.Any = None,
) -> bytes:
    if len(wrapping_key) not in [16, 24, 32]:
        raise ValueError("The wrapping key must be a valid AES key length")

    if len(key_to_wrap) < 16:
        raise ValueError("The key to wrap must be at least 16 bytes")

    if len(key_to_wrap) % 8 != 0:
        raise ValueError("The key to wrap must be a multiple of 8 bytes")

    a = b"\xa6\xa6\xa6\xa6\xa6\xa6\xa6\xa6"
    r = [key_to_wrap[i : i + 8] for i in range(0, len(key_to_wrap), 8)]
    return _wrap_core(wrapping_key, a, r)


def _unwrap_core(
    wrapping_key: bytes,
    a: bytes,
    r: list[bytes],
) -> tuple[bytes, list[bytes]]:
    # Implement RFC 3394 Key Unwrap - 2.2.2 (index method)
    decryptor = Cipher(AES(wrapping_key), ECB()).decryptor()
    n = len(r)
    for j in reversed(range(6)):
        for i in reversed(range(n)):
            atr = (
                int.from_bytes(a, byteorder="big") ^ ((n * j) + i + 1)
            ).to_bytes(length=8, byteorder="big") + r[i]
            # every decryption operation is a discrete 16 byte chunk so
            # it is safe to reuse the decryptor for the entire operation
            b = decryptor.update(atr)
            a = b[:8]
            r[i] = b[-8:]

    assert decryptor.finalize() == b""
    return a, r


def aes_key_wrap_with_padding(
    wrapping_key: bytes,
    key_to_wrap: bytes,
    backend: typing.Any = None,
) -> bytes:
    if len(wrapping_key) not in [16, 24, 32]:
        raise ValueError("The wrapping key must be a valid AES key length")

    aiv = b"\xa6\x59\x59\xa6" + len(key_to_wrap).to_bytes(
        length=4, byteorder="big"
    )
    # pad the key to wrap if necessary
    pad = (8 - (len(key_to_wrap) % 8)) % 8
    key_to_wrap = key_to_wrap + b"\x00" * pad
    if len(key_to_wrap) == 8:
        # RFC 5649 - 4.1 - exactly 8 octets after padding
        encryptor = Cipher(AES(wrapping_key), ECB()).encryptor()
        b = encryptor.update(aiv + key_to_wrap)
        assert encryptor.finalize() == b""
        return b
    else:
        r = [key_to_wrap[i : i + 8] for i in range(0, len(key_to_wrap), 8)]
        return _wrap_core(wrapping_key, aiv, r)


def aes_key_unwrap_with_padding(
    wrapping_key: bytes,
    wrapped_key: bytes,
    backend: typing.Any = None,
) -> bytes:
    if len(wrapped_key) < 16:
        raise InvalidUnwrap("Must be at least 16 bytes")

    if len(wrapping_key) not in [16, 24, 32]:
        raise ValueError("The wrapping key must be a valid AES key length")

    if len(wrapped_key) == 16:
        # RFC 5649 - 4.2 - exactly two 64-bit blocks
        decryptor = Cipher(AES(wrapping_key), ECB()).decryptor()
        out = decryptor.update(wrapped_key)
        assert decryptor.finalize() == b""
        a = out[:8]
        data = out[8:]
        n = 1
    else:
        r = [wrapped_key[i : i + 8] for i in range(0, len(wrapped_key), 8)]
        encrypted_aiv = r.pop(0)
        n = len(r)
        a, r = _unwrap_core(wrapping_key, encrypted_aiv, r)
        data = b"".join(r)

    # 1) Check that MSB(32,A) = A65959A6.
    # 2) Check that 8*(n-1) < LSB(32,A) <= 8*n.  If so, let
    #    MLI = LSB(32,A).
    # 3) Let b = (8*n)-MLI, and then check that the rightmost b octets of
    #    the output data are zero.
    mli = int.from_bytes(a[4:], byteorder="big")
    b = (8 * n) - mli
    if (
        not bytes_eq(a[:4], b"\xa6\x59\x59\xa6")
        or not 8 * (n - 1) < mli <= 8 * n
        or (b != 0 and not bytes_eq(data[-b:], b"\x00" * b))
    ):
        raise InvalidUnwrap()

    if b == 0:
        return data
    else:
        return data[:-b]


def aes_key_unwrap(
    wrapping_key: bytes,
    wrapped_key: bytes,
    backend: typing.Any = None,
) -> bytes:
    if len(wrapped_key) < 24:
        raise InvalidUnwrap("Must be at least 24 bytes")

    if len(wrapped_key) % 8 != 0:
        raise InvalidUnwrap("The wrapped key must be a multiple of 8 bytes")

    if len(wrapping_key) not in [16, 24, 32]:
        raise ValueError("The wrapping key must be a valid AES key length")

    aiv = b"\xa6\xa6\xa6\xa6\xa6\xa6\xa6\xa6"
    r = [wrapped_key[i : i + 8] for i in range(0, len(wrapped_key), 8)]
    a = r.pop(0)
    a, r = _unwrap_core(wrapping_key, a, r)
    if not bytes_eq(a, aiv):
        raise InvalidUnwrap()

    return b"".join(r)


class InvalidUnwrap(Exception):
    pass



from production.model import PicklistCleanerManagerConfig
from typing import List


class ProductionConfig(object):
    def __init__(
        self,                                   # type: ProductionConfig
        boxes_config,                           # type: List[object]
        branches_config,                        # type: List[object]
        views_config,                           # type: object
        picklist_database_manager_config,       # type: PicklistCleanerManagerConfig
    ):
        # type: (...) -> None

        self.boxes_config = boxes_config
        self.branches_config = branches_config
        self.views_config = views_config
        self.picklist_database_manager_config = picklist_database_manager_config



# testing/provision.py
# Copyright (C) 2005-2025 the SQLAlchemy authors and contributors
# <see AUTHORS file>
#
# This module is part of SQLAlchemy and is released under
# the MIT License: https://www.opensource.org/licenses/mit-license.php
# mypy: ignore-errors

from __future__ import annotations

import collections
import logging

from . import config
from . import engines
from . import util
from .. import exc
from .. import inspect
from ..engine import url as sa_url
from ..sql import ddl
from ..sql import schema


log = logging.getLogger(__name__)

FOLLOWER_IDENT = None


class register:
    def __init__(self, decorator=None):
        self.fns = {}
        self.decorator = decorator

    @classmethod
    def init(cls, fn):
        return register().for_db("*")(fn)

    @classmethod
    def init_decorator(cls, decorator):
        return register(decorator).for_db("*")

    def for_db(self, *dbnames):
        def decorate(fn):
            if self.decorator:
                fn = self.decorator(fn)
            for dbname in dbnames:
                self.fns[dbname] = fn
            return self

        return decorate

    def __call__(self, cfg, *arg, **kw):
        if isinstance(cfg, str):
            url = sa_url.make_url(cfg)
        elif isinstance(cfg, sa_url.URL):
            url = cfg
        else:
            url = cfg.db.url
        backend = url.get_backend_name()
        if backend in self.fns:
            return self.fns[backend](cfg, *arg, **kw)
        else:
            return self.fns["*"](cfg, *arg, **kw)


def create_follower_db(follower_ident):
    for cfg in _configs_for_db_operation():
        log.info("CREATE database %s, URI %r", follower_ident, cfg.db.url)
        create_db(cfg, cfg.db, follower_ident)


def setup_config(db_url, options, file_config, follower_ident):
    # load the dialect, which should also have it set up its provision
    # hooks

    dialect = sa_url.make_url(db_url).get_dialect()

    dialect.load_provisioning()

    if follower_ident:
        db_url = follower_url_from_main(db_url, follower_ident)
    db_opts = {}
    update_db_opts(db_url, db_opts, options)
    db_opts["scope"] = "global"
    eng = engines.testing_engine(db_url, db_opts)
    post_configure_engine(db_url, eng, follower_ident)
    eng.connect().close()

    cfg = config.Config.register(eng, db_opts, options, file_config)

    # a symbolic name that tests can use if they need to disambiguate
    # names across databases
    if follower_ident:
        config.ident = follower_ident

    if follower_ident:
        configure_follower(cfg, follower_ident)
    return cfg


def drop_follower_db(follower_ident):
    for cfg in _configs_for_db_operation():
        log.info("DROP database %s, URI %r", follower_ident, cfg.db.url)
        drop_db(cfg, cfg.db, follower_ident)


def generate_db_urls(db_urls, extra_drivers):
    """Generate a set of URLs to test given configured URLs plus additional
    driver names.

    Given:

    .. sourcecode:: text

        --dburi postgresql://db1  \
        --dburi postgresql://db2  \
        --dburi postgresql://db2  \
        --dbdriver=psycopg2 --dbdriver=asyncpg?async_fallback=true

    Noting that the default postgresql driver is psycopg2,  the output
    would be:

    .. sourcecode:: text

        postgresql+psycopg2://db1
        postgresql+asyncpg://db1
        postgresql+psycopg2://db2
        postgresql+psycopg2://db3

    That is, for the driver in a --dburi, we want to keep that and use that
    driver for each URL it's part of .   For a driver that is only
    in --dbdrivers, we want to use it just once for one of the URLs.
    for a driver that is both coming from --dburi as well as --dbdrivers,
    we want to keep it in that dburi.

    Driver specific query options can be specified by added them to the
    driver name. For example, to enable the async fallback option for
    asyncpg::

    .. sourcecode:: text

        --dburi postgresql://db1  \
        --dbdriver=asyncpg?async_fallback=true

    """
    urls = set()

    backend_to_driver_we_already_have = collections.defaultdict(set)

    urls_plus_dialects = [
        (url_obj, url_obj.get_dialect())
        for url_obj in [sa_url.make_url(db_url) for db_url in db_urls]
    ]

    for url_obj, dialect in urls_plus_dialects:
        # use get_driver_name instead of dialect.driver to account for
        # "_async" virtual drivers like oracledb and psycopg
        driver_name = url_obj.get_driver_name()
        backend_to_driver_we_already_have[dialect.name].add(driver_name)

    backend_to_driver_we_need = {}

    for url_obj, dialect in urls_plus_dialects:
        backend = dialect.name
        dialect.load_provisioning()

        if backend not in backend_to_driver_we_need:
            backend_to_driver_we_need[backend] = extra_per_backend = set(
                extra_drivers
            ).difference(backend_to_driver_we_already_have[backend])
        else:
            extra_per_backend = backend_to_driver_we_need[backend]

        for driver_url in _generate_driver_urls(url_obj, extra_per_backend):
            if driver_url in urls:
                continue
            urls.add(driver_url)
            yield driver_url


def _generate_driver_urls(url, extra_drivers):
    main_driver = url.get_driver_name()
    extra_drivers.discard(main_driver)

    url = generate_driver_url(url, main_driver, "")
    yield url

    for drv in list(extra_drivers):
        if "?" in drv:
            driver_only, query_str = drv.split("?", 1)

        else:
            driver_only = drv
            query_str = None

        new_url = generate_driver_url(url, driver_only, query_str)
        if new_url:
            extra_drivers.remove(drv)

            yield new_url


@register.init
def generate_driver_url(url, driver, query_str):
    backend = url.get_backend_name()

    new_url = url.set(
        drivername="%s+%s" % (backend, driver),
    )
    if query_str:
        new_url = new_url.update_query_string(query_str)

    try:
        new_url.get_dialect()
    except exc.NoSuchModuleError:
        return None
    else:
        return new_url


def _configs_for_db_operation():
    hosts = set()

    for cfg in config.Config.all_configs():
        cfg.db.dispose()

    for cfg in config.Config.all_configs():
        url = cfg.db.url
        backend = url.get_backend_name()
        host_conf = (backend, url.username, url.host, url.database)

        if host_conf not in hosts:
            yield cfg
            hosts.add(host_conf)

    for cfg in config.Config.all_configs():
        cfg.db.dispose()


@register.init
def drop_all_schema_objects_pre_tables(cfg, eng):
    pass


@register.init
def drop_all_schema_objects_post_tables(cfg, eng):
    pass


def drop_all_schema_objects(cfg, eng):
    drop_all_schema_objects_pre_tables(cfg, eng)

    drop_views(cfg, eng)

    if config.requirements.materialized_views.enabled:
        drop_materialized_views(cfg, eng)

    inspector = inspect(eng)

    consider_schemas = (None,)
    if config.requirements.schemas.enabled_for_config(cfg):
        consider_schemas += (cfg.test_schema, cfg.test_schema_2)
    util.drop_all_tables(eng, inspector, consider_schemas=consider_schemas)

    drop_all_schema_objects_post_tables(cfg, eng)

    if config.requirements.sequences.enabled_for_config(cfg):
        with eng.begin() as conn:
            for seq in inspector.get_sequence_names():
                conn.execute(ddl.DropSequence(schema.Sequence(seq)))
            if config.requirements.schemas.enabled_for_config(cfg):
                for schema_name in [cfg.test_schema, cfg.test_schema_2]:
                    for seq in inspector.get_sequence_names(
                        schema=schema_name
                    ):
                        conn.execute(
                            ddl.DropSequence(
                                schema.Sequence(seq, schema=schema_name)
                            )
                        )


def drop_views(cfg, eng):
    inspector = inspect(eng)

    try:
        view_names = inspector.get_view_names()
    except NotImplementedError:
        pass
    else:
        with eng.begin() as conn:
            for vname in view_names:
                conn.execute(
                    ddl._DropView(schema.Table(vname, schema.MetaData()))
                )

    if config.requirements.schemas.enabled_for_config(cfg):
        try:
            view_names = inspector.get_view_names(schema=cfg.test_schema)
        except NotImplementedError:
            pass
        else:
            with eng.begin() as conn:
                for vname in view_names:
                    conn.execute(
                        ddl._DropView(
                            schema.Table(
                                vname,
                                schema.MetaData(),
                                schema=cfg.test_schema,
                            )
                        )
                    )


def drop_materialized_views(cfg, eng):
    inspector = inspect(eng)

    mview_names = inspector.get_materialized_view_names()

    with eng.begin() as conn:
        for vname in mview_names:
            conn.exec_driver_sql(f"DROP MATERIALIZED VIEW {vname}")

    if config.requirements.schemas.enabled_for_config(cfg):
        mview_names = inspector.get_materialized_view_names(
            schema=cfg.test_schema
        )
        with eng.begin() as conn:
            for vname in mview_names:
                conn.exec_driver_sql(
                    f"DROP MATERIALIZED VIEW {cfg.test_schema}.{vname}"
                )


@register.init
def create_db(cfg, eng, ident):
    """Dynamically create a database for testing.

    Used when a test run will employ multiple processes, e.g., when run
    via `tox` or `pytest -n4`.
    """
    raise NotImplementedError(
        "no DB creation routine for cfg: %s" % (eng.url,)
    )


@register.init
def drop_db(cfg, eng, ident):
    """Drop a database that we dynamically created for testing."""
    raise NotImplementedError("no DB drop routine for cfg: %s" % (eng.url,))


def _adapt_update_db_opts(fn):
    insp = util.inspect_getfullargspec(fn)
    if len(insp.args) == 3:
        return fn
    else:
        return lambda db_url, db_opts, _options: fn(db_url, db_opts)


@register.init_decorator(_adapt_update_db_opts)
def update_db_opts(db_url, db_opts, options):
    """Set database options (db_opts) for a test database that we created."""


@register.init
def post_configure_engine(url, engine, follower_ident):
    """Perform extra steps after configuring an engine for testing.

    (For the internal dialects, currently only used by sqlite, oracle, mssql)
    """


@register.init
def follower_url_from_main(url, ident):
    """Create a connection URL for a dynamically-created test database.

    :param url: the connection URL specified when the test run was invoked
    :param ident: the pytest-xdist "worker identifier" to be used as the
                  database name
    """
    url = sa_url.make_url(url)
    return url.set(database=ident)


@register.init
def configure_follower(cfg, ident):
    """Create dialect-specific config settings for a follower database."""
    pass


@register.init
def run_reap_dbs(url, ident):
    """Remove databases that were created during the test process, after the
    process has ended.

    This is an optional step that is invoked for certain backends that do not
    reliably release locks on the database as long as a process is still in
    use. For the internal dialects, this is currently only necessary for
    mssql and oracle.
    """


def reap_dbs(idents_file):
    log.info("Reaping databases...")

    urls = collections.defaultdict(set)
    idents = collections.defaultdict(set)
    dialects = {}

    with open(idents_file) as file_:
        for line in file_:
            line = line.strip()
            db_name, db_url = line.split(" ")
            url_obj = sa_url.make_url(db_url)
            if db_name not in dialects:
                dialects[db_name] = url_obj.get_dialect()
                dialects[db_name].load_provisioning()
            url_key = (url_obj.get_backend_name(), url_obj.host)
            urls[url_key].add(db_url)
            idents[url_key].add(db_name)

    for url_key in urls:
        url = list(urls[url_key])[0]
        ident = idents[url_key]
        run_reap_dbs(url, ident)


@register.init
def temp_table_keyword_args(cfg, eng):
    """Specify keyword arguments for creating a temporary Table.

    Dialect-specific implementations of this method will return the
    kwargs that are passed to the Table method when creating a temporary
    table for testing, e.g., in the define_temp_tables method of the
    ComponentReflectionTest class in suite/test_reflection.py
    """
    raise NotImplementedError(
        "no temp table keyword args routine for cfg: %s" % (eng.url,)
    )


@register.init
def prepare_for_drop_tables(config, connection):
    pass


@register.init
def stop_test_class_outside_fixtures(config, db, testcls):
    pass


@register.init
def get_temp_table_name(cfg, eng, base_name):
    """Specify table name for creating a temporary Table.

    Dialect-specific implementations of this method will return the
    name to use when creating a temporary table for testing,
    e.g., in the define_temp_tables method of the
    ComponentReflectionTest class in suite/test_reflection.py

    Default to just the base name since that's what most dialects will
    use. The mssql dialect's implementation will need a "#" prepended.
    """
    return base_name


@register.init
def set_default_schema_on_connection(cfg, dbapi_connection, schema_name):
    raise NotImplementedError(
        "backend does not implement a schema name set function: %s"
        % (cfg.db.url,)
    )


@register.init
def upsert(
    cfg, table, returning, *, set_lambda=None, sort_by_parameter_order=False
):
    """return the backends insert..on conflict / on dupe etc. construct.

    while we should add a backend-neutral upsert construct as well, such as
    insert().upsert(), it's important that we continue to test the
    backend-specific insert() constructs since if we do implement
    insert().upsert(), that would be using a different codepath for the things
    we need to test like insertmanyvalues, etc.

    """
    raise NotImplementedError(
        f"backend does not include an upsert implementation: {cfg.db.url}"
    )


@register.init
def normalize_sequence(cfg, sequence):
    """Normalize sequence parameters for dialect that don't start with 1
    by default.

    The default implementation does nothing
    """
    return sequence



import os.path
import pathlib
import unittest

from importlib import import_module
from importlib_resources.readers import MultiplexedPath, NamespaceReader

from . import util


class MultiplexedPathTest(util.DiskSetup, unittest.TestCase):
    MODULE = 'namespacedata01'

    def setUp(self):
        super().setUp()
        self.folder = pathlib.Path(self.data.__path__[0])
        self.data01 = pathlib.Path(self.load_fixture('data01').__file__).parent
        self.data02 = pathlib.Path(self.load_fixture('data02').__file__).parent

    def test_init_no_paths(self):
        with self.assertRaises(FileNotFoundError):
            MultiplexedPath()

    def test_init_file(self):
        with self.assertRaises(NotADirectoryError):
            MultiplexedPath(self.folder / 'binary.file')

    def test_iterdir(self):
        contents = {path.name for path in MultiplexedPath(self.folder).iterdir()}
        try:
            contents.remove('__pycache__')
        except (KeyError, ValueError):
            pass
        self.assertEqual(
            contents, {'subdirectory', 'binary.file', 'utf-16.file', 'utf-8.file'}
        )

    def test_iterdir_duplicate(self):
        contents = {
            path.name for path in MultiplexedPath(self.folder, self.data01).iterdir()
        }
        for remove in ('__pycache__', '__init__.pyc'):
            try:
                contents.remove(remove)
            except (KeyError, ValueError):
                pass
        self.assertEqual(
            contents,
            {'__init__.py', 'binary.file', 'subdirectory', 'utf-16.file', 'utf-8.file'},
        )

    def test_is_dir(self):
        self.assertEqual(MultiplexedPath(self.folder).is_dir(), True)

    def test_is_file(self):
        self.assertEqual(MultiplexedPath(self.folder).is_file(), False)

    def test_open_file(self):
        path = MultiplexedPath(self.folder)
        with self.assertRaises(FileNotFoundError):
            path.read_bytes()
        with self.assertRaises(FileNotFoundError):
            path.read_text()
        with self.assertRaises(FileNotFoundError):
            path.open()

    def test_join_path(self):
        prefix = str(self.folder.parent)
        path = MultiplexedPath(self.folder, self.data01)
        self.assertEqual(
            str(path.joinpath('binary.file'))[len(prefix) + 1 :],
            os.path.join('namespacedata01', 'binary.file'),
        )
        sub = path.joinpath('subdirectory')
        assert isinstance(sub, MultiplexedPath)
        assert 'namespacedata01' in str(sub)
        assert 'data01' in str(sub)
        self.assertEqual(
            str(path.joinpath('imaginary'))[len(prefix) + 1 :],
            os.path.join('namespacedata01', 'imaginary'),
        )
        self.assertEqual(path.joinpath(), path)

    def test_join_path_compound(self):
        path = MultiplexedPath(self.folder)
        assert not path.joinpath('imaginary/foo.py').exists()

    def test_join_path_common_subdir(self):
        prefix = str(self.data02.parent)
        path = MultiplexedPath(self.data01, self.data02)
        self.assertIsInstance(path.joinpath('subdirectory'), MultiplexedPath)
        self.assertEqual(
            str(path.joinpath('subdirectory', 'subsubdir'))[len(prefix) + 1 :],
            os.path.join('data02', 'subdirectory', 'subsubdir'),
        )

    def test_repr(self):
        self.assertEqual(
            repr(MultiplexedPath(self.folder)),
            f"MultiplexedPath('{self.folder}')",
        )

    def test_name(self):
        self.assertEqual(
            MultiplexedPath(self.folder).name,
            os.path.basename(self.folder),
        )


class NamespaceReaderTest(util.DiskSetup, unittest.TestCase):
    MODULE = 'namespacedata01'

    def test_init_error(self):
        with self.assertRaises(ValueError):
            NamespaceReader(['path1', 'path2'])

    def test_resource_path(self):
        namespacedata01 = import_module('namespacedata01')
        reader = NamespaceReader(namespacedata01.__spec__.submodule_search_locations)

        root = self.data.__path__[0]
        self.assertEqual(
            reader.resource_path('binary.file'), os.path.join(root, 'binary.file')
        )
        self.assertEqual(
            reader.resource_path('imaginary'), os.path.join(root, 'imaginary')
        )

    def test_files(self):
        reader = NamespaceReader(self.data.__spec__.submodule_search_locations)
        root = self.data.__path__[0]
        self.assertIsInstance(reader.files(), MultiplexedPath)
        self.assertEqual(repr(reader.files()), f"MultiplexedPath('{root}')")


if __name__ == '__main__':
    unittest.main()



# Protocol Buffers - Google's data interchange format
# Copyright 2008 Google Inc.  All rights reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file or at
# https://developers.google.com/open-source/licenses/bsd

"""Contains the Timestamp helper APIs."""

import datetime
from typing import Optional

from google.protobuf.timestamp_pb2 import Timestamp


def from_json_string(value: str) -> Timestamp:
  """Parse a RFC 3339 date string format to Timestamp.

  Args:
    value: A date string. Any fractional digits (or none) and any offset are
      accepted as long as they fit into nano-seconds precision. Example of
      accepted format: '1972-01-01T10:00:20.021-05:00'

  Raises:
    ValueError: On parsing problems.
  """
  timestamp = Timestamp()
  timestamp.FromJsonString(value)
  return timestamp


def from_microseconds(micros: float) -> Timestamp:
  """Converts microseconds since epoch to Timestamp."""
  timestamp = Timestamp()
  timestamp.FromMicroseconds(micros)
  return timestamp


def from_milliseconds(millis: float) -> Timestamp:
  """Converts milliseconds since epoch to Timestamp."""
  timestamp = Timestamp()
  timestamp.FromMilliseconds(millis)
  return timestamp


def from_nanoseconds(nanos: float) -> Timestamp:
  """Converts nanoseconds since epoch to Timestamp."""
  timestamp = Timestamp()
  timestamp.FromNanoseconds(nanos)
  return timestamp


def from_seconds(seconds: float) -> Timestamp:
  """Converts seconds since epoch to Timestamp."""
  timestamp = Timestamp()
  timestamp.FromSeconds(seconds)
  return timestamp


def from_current_time() -> Timestamp:
  """Converts the current UTC to Timestamp."""
  timestamp = Timestamp()
  timestamp.FromDatetime(datetime.datetime.now(tz=datetime.timezone.utc))
  return timestamp


def to_json_string(ts: Timestamp) -> str:
  """Converts Timestamp to RFC 3339 date string format.

  Returns:
    A string converted from timestamp. The string is always Z-normalized
    and uses 3, 6 or 9 fractional digits as required to represent the
    exact time. Example of the return format: '1972-01-01T10:00:20.021Z'
  """
  return ts.ToJsonString()


def to_microseconds(ts: Timestamp) -> int:
  """Converts Timestamp to microseconds since epoch."""
  return ts.ToMicroseconds()


def to_milliseconds(ts: Timestamp) -> int:
  """Converts Timestamp to milliseconds since epoch."""
  return ts.ToMilliseconds()


def to_nanoseconds(ts: Timestamp) -> int:
  """Converts Timestamp to nanoseconds since epoch."""
  return ts.ToNanoseconds()


def to_seconds(ts: Timestamp) -> int:
  """Converts Timestamp to seconds since epoch."""
  return ts.ToSeconds()


def to_datetime(
    ts: Timestamp, tz: Optional[datetime.tzinfo] = None
) -> datetime.datetime:
  """Converts Timestamp to a datetime.

  Args:
    tz: A datetime.tzinfo subclass; defaults to None.

  Returns:
    If tzinfo is None, returns a timezone-naive UTC datetime (with no timezone
    information, i.e. not aware that it's UTC).

    Otherwise, returns a timezone-aware datetime in the input timezone.
  """
  return ts.ToDatetime(tzinfo=tz)



import logging

from table_manager_api.dto.exception import (
    TableError,
    TabNameAlreadyExists, InvalidTabName,
)
from typing import (
    Any,
    List,
    Optional,
    Tuple,
    Union,
)
from xml.etree import cElementTree as eTree

from mw_helper import (
    show_message_dialog,
    show_message_options_dialog,
)
from pos_model import (
    TableStatus,
    TableOperationMode,
)
from sysactions import (
    StopAction,
    action,
    show_confirmation,
    get_model,
)

from api.table import get_table_service
from actions.config import get_tab_enabled
from actions.table.create_tab import get_tab_info
from actions.table.unlink_tabs import unlink_multiple_tabs
from application.domain import Table

from msgbusboundary import MsgBusTableService as tableService
from tablemgrapi import (
    TableServiceException,
    get_pos_ts,
)

logger = logging.getLogger("TableActions")


@action
def deselect_table(
    pos_id,                     # type: Union[str, int]
    pre_selected_option=False,  # type: bool
    do_raise=True               # type: bool
):
    # type: (...) -> bool

    try:
        from table_actions import (
            has_table_service,
            create_table_modified_event,
        )

        if not has_table_service():
            return True

        tables = list(tableService().list_tables(pos_id=pos_id))
        selected_table = tableService().get_selected_table(
            pos_id=pos_id,
            tables=tables,
        )
        if not selected_table:
            create_table_modified_event(
                pos_id=pos_id,
                table_picture="0",
            )
            return True

        table_picture = eTree.XML(tableService().get_table_picture(
            pos_id=pos_id,
            table_id=selected_table.id,
        ))
        is_a_tab = tableService().is_a_tab(
            pos_id=pos_id,
            table_id=selected_table.id,
            table_picture=table_picture,
        )

        if is_a_tab:
            table_picture, tables = confirm_operation_on_mode_tabs_with_tabs(
                pos_id=pos_id,
                table_picture=table_picture,
                tables=tables,
            )

        try:
            if is_a_tab and not pre_selected_option:
                finish_action, element = _deselect_tab_action(
                    pos_id=pos_id,
                    table_picture=table_picture,
                    do_raise=do_raise,
                )
                if finish_action:
                    return element

            check_if_tab_name_is_required(
                pos_id=pos_id,
                table_id=table_picture.get("id"),
                table_picture=table_picture,
            )

        except TableServiceException:
            logger.exception("Error getting table picture")

        try:
            for table in tables:
                if pos_id == table.current_pos_id:
                    tableService().store_service(
                        pos_id=pos_id,
                        table_id=str(table.id),
                    )
        except TableServiceException:
            show_message_dialog(
                pos_id=pos_id,
                message="$COULD_NOT_DESELECT_TABLE",
            )
            logger.exception("Error getting table picture")

            return False

        create_table_modified_event(
            pos_id=pos_id,
            table_picture="0",
        )

        return True

    except StopAction:
        return False


def check_can_return_to_original_table(
    pos_id,                 # type: str
    slice_source_id,        # type: str
    current_table_picture,  # type: eTree.Element
):
    # type: (...) -> bool

    if not slice_source_id:
        return False

    source_table_picture = eTree.XML(tableService().get_table_picture(
        pos_id=pos_id,
        table_id=slice_source_id,
    ))
    status = int(source_table_picture.get("status"))
    if status in [TableStatus.CLOSED, TableStatus.ABANDONED]:
        return False

    if status == TableStatus.AVAILABLE:
        return True

    return _check_if_source_table_still_has_the_same_service(
        current_table_picture=current_table_picture,
        source_table_picture=source_table_picture,
    )


def _check_if_source_table_still_has_the_same_service(
    current_table_picture,  # type: eTree.Element
    source_table_picture,   # type: eTree.Element
):
    # type: (...) -> bool

    return current_table_picture.get("sourceServiceId") == source_table_picture.get("serviceId")


def check_if_tab_name_is_required(
    pos_id,                 # type: str
    table_id,               # type: str
    table_picture=None,     # type: Optional[eTree.XML]
):
    # type: (...) -> bool

    if not table_picture:
        table_picture = eTree.fromstring(tableService().get_table_info(
            pos_id=pos_id,
            table_id=table_id,
        ))

    table_description = table_picture.get("typeDescr")

    if table_description == "Tab":
        tab_name = table_picture.get("name")
        if tab_name:
            return False

        customer_data = get_tab_info(pos_id=pos_id)
        if not customer_data:
            raise StopAction()

        table_picture, changed_user = check_tab_user_sector(
            table_picture=table_picture,
            pos_id=pos_id,
        )
        if table_picture is None:
            raise StopAction()

        tab_name = customer_data.pop("TAB_NAME", None)
        if not tab_name:
            raise StopAction()

        try:
            get_table_service(pos_id=pos_id).create_tab_from_split_service(
                table_id=table_id,
                table_name=tab_name,
                properties=customer_data,
            )
        except TableError as ex:
            message = "$CREATE_TAB_FROM_SPLIT_SERVICE_ERROR"
            if isinstance(ex, TabNameAlreadyExists):
                message = "$TAB_ALREADY_EXISTS"
            elif isinstance(ex, InvalidTabName):
                message = "$EXISTS_AS_A_TABLE"

            show_message_dialog(
                pos_id=pos_id,
                message=message,
            )
            message = "Error trying to set info tab_name: {}, properties: {} into table_id: {}"
            logger.exception(message.format(tab_name, customer_data, table_id))
            raise StopAction()

        return changed_user


def check_tab_user_sector(
    table_picture,  # type: Union[str, eTree.XML]
    pos_id,         # type: str
    reopen=False,   # type: bool
):
    # type: (...) -> (Tuple[Union[str, eTree.XML, None]], bool)

    from table_actions import table_operation_mode

    sector_mode = table_operation_mode in [TableOperationMode.SHARED_SECTOR, TableOperationMode.INDIVIDUAL_SECTOR]
    if not sector_mode:
        return table_picture, False

    model = get_model(posid=pos_id)
    pos_ts = get_pos_ts(model=model)
    table_picture_xml = eTree.XML(table_picture) if isinstance(table_picture, str) else table_picture
    table_user_id = int(table_picture_xml.get("userId"))
    table_id = table_picture_xml.get("id")

    if reopen:
        changed_user = check_if_tab_name_is_required(
            pos_id=pos_id,
            table_id=table_id,
        )
    else:

        from table_actions import show_list_operators

        new_operator = show_list_operators(
            pos_id=pos_id,
            sector="$TABS",
            table_user=table_user_id,
        )
        if new_operator[0] < 0:
            return None, False

        new_operator = new_operator[1][new_operator[0]]

        pos_ts.changeTableUser(
            user_id=new_operator,
            table_id=table_id,
        )

        changed_user = True if new_operator != table_user_id and sector_mode else False

    table_picture = tableService().get_table_picture(
        pos_id=pos_id,
        table_id=table_id,
    )

    return table_picture, changed_user


def _deselect_tab_action(
    pos_id,             # type: str
    table_picture,      # type: eTree.XML
    do_raise,           # type: bool
):
    # type: (...) -> Tuple[bool, Any]

    slice_source_id = table_picture.get("sliceSourceTableId")
    tab_name = table_picture.get("name")
    can_return_to_original_table = check_can_return_to_original_table(
        pos_id=pos_id,
        slice_source_id=slice_source_id,
        current_table_picture=table_picture,
    )

    if slice_source_id and not tab_name:
        option_back_from_source_table = 0
        if get_tab_enabled():
            options = "$BACK_FROM_SOURCE_TABLE|$DESELECT_TABLE|$CANCEL"
            option_cancel = 2
        else:
            options = "$BACK_FROM_SOURCE_TABLE|$CANCEL"
            option_cancel = 1

        ret = show_message_options_dialog(
            pos_id=pos_id,
            buttons=options,
            message="$CHOOSE_DESELECT_TYPE",
        )
        if ret == option_back_from_source_table:

            from table_actions import do_transfer_table

            return True, do_transfer_table(
                pos_id=pos_id,
                parent_table_id=table_picture.get("id"),
                destination_table_id=slice_source_id if can_return_to_original_table else None,
                select_destination_table=False,
                ignore_totaled_orders=True,
            )

        if ret == option_cancel:
            if not do_raise:
                return True, False

            raise StopAction()

    return False, False


def confirm_operation_on_mode_tabs_with_tabs(
    pos_id,         # type: str
    table_picture,  # type: eTree.Element
    tables=None,    # type: Optional[List[Table]]
):
    # type: (...) -> Tuple[eTree.Element, List[Table]]

    services_children = table_picture.findall(".//ServiceChildren/ServiceChild")
    if len(services_children) == 0:
        return table_picture, tables

    from table_actions import check_if_table_have_tenders
    has_tenders = check_if_table_have_tenders(
        pos_id=pos_id,
        table_picture=table_picture,
        message="$TABLE_HAS_TENDERS",
    )
    if has_tenders:
        raise StopAction()

    confirm = True
    if len(services_children) > 1:
        confirm = show_confirmation(
            posid=pos_id,
            message="$TO_PROCEED_THE_LINKED_TABS_MUST_BE_SEPARATED",
        )

    if not confirm:
        raise StopAction()

    to_unlink = []
    for child in services_children:
        to_unlink.append(child.get("tableId"))

    table_picture_dto = unlink_multiple_tabs(
        pos_id=pos_id,
        current_tab_id=table_picture.get("id"),
        tabs_to_unlink=to_unlink,
    )

    if table_picture_dto is None:
        raise StopAction()

    tables = list(tableService().list_tables(pos_id=pos_id))
    return table_picture_dto.to_xml(), tables



import json
from logging import Logger

from msgbus import (
    MBEasyContext,
    TK_SYS_ACK
)
from messagebus import (
    TokenCreator,
    TokenPriority,
)

MSGGRP_DEVICE_MANAGER = "D"
TK_GET_USED_SERVICE = TokenCreator.create_token(TokenPriority.low, MSGGRP_DEVICE_MANAGER, "1")


class DeviceManagerService(object):
    def __init__(
        self,
        mb_context,  # type: MBEasyContext
        logger       # type: Logger
    ):
        # type: (...) -> None

        self.mb_context = mb_context
        self.logger = logger

    def get_available_kds(
        self,
        kds_id
    ):
        # type: (...) -> int

        try:
            data = json.dumps({
                "kdsId": kds_id,
                "deviceType": "display",
            })

            result = self.mb_context.MB_EasySendMessage(
                dest_name="DeviceManager",
                token=TK_GET_USED_SERVICE,
                data=data,
                timeout=5000000
            )
            if result.token == TK_SYS_ACK:
                return int(result.data.split("KDS")[-1])
        except ValueError:
            pass

        except (Exception,):
            self.logger.exception("Error getting fallback KDS for KDS Id {}".format(kds_id))

        return int(kds_id)
